from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, Any
if TYPE_CHECKING:
    from pyvol_terminal.gl_3D_graphing.graphics_items.GL3DViewBox import GL3DViewBox
    from pyvol_terminal.gl_3D_graphing.widgets.GL3DViewWidget import GL3DViewWidget

import pyqtgraph.opengl as gl

from pyqtgraph.opengl import GLGraphicsItem, GLViewWidget
from pyqtgraph import opengl
from PySide6 import QtWidgets, QtCore, QtGui
import numpy as np
from OpenGL import GL
from dataclasses import dataclass, field, InitVar, asdict
import weakref
from pyqtgraph.Qt import isQObjectAlive
from pyvol_terminal.gl_3D_graphing.meta import QABCMeta, abc
import pyqtgraph as pg
from pyqtgraph import getConfigOption
from .GL3DTextItem import GL3DTextItem
from .GL3DGraphicsItems import GL3DLinePlotDataItem, GL3DTextItem
from .GL3DGraphicsItemMixin import GL3DGraphicsItemMixin
import collections
from . import validationHelpers


import math
from pprint import pprint

from pyqtgraph.Qt import QtGui, QT_LIB
import importlib

if QT_LIB in ["PyQt5", "PySide2"]:
    QtOpenGL = QtGui
else:
    QtOpenGL = importlib.import_module(f"{QT_LIB}.QtOpenGL")




class _TickLineset(opengl.GLLinePlotItem):
    def __init__(self,
                 direction: int,
                 baseCoordTicks: List[List[np.ndarray]]=None, 
                 baseCoordGrid: Tuple[np.ndarray, np.ndarray]=None,
                 borderCoordinates:  Tuple[List[float], List[float]]=None,
                 color: Tuple[float, ...]=getConfigOption('foreground'),
                 width: float=1,
                 showGrid: bool=True,
                 showBorder: bool=False,
                 parentItem=None,
                 **kwargs
                 ):
        
        mode = "lines"    # Keep mode lines for optimized painting for grouped line items 
        
        super().__init__(parentItem=parentItem, color=color, width=width, mode=mode, **kwargs)
        self.direction=direction
        self.baseCoordTicks=baseCoordTicks
        self.baseCoordGrid=baseCoordGrid
        self.tickVector=None
        self.gapOffset=None
        self._border_flag=showBorder


        self.border_cells=None
        self.showGrid=showGrid
        self.base_tick_coords: List[List[float]]=None
        self.base_grid_coords: List[List[float]]=None
        self.endGridCoord=0
        
        if self._border_flag:
            self.border_cells = borderCoordinates

        
    def updateTickPositions(self, tickLevelsWorld):
        tick_arr = []
        t_T=None
        tick_pos=None

        
        for jdx, tick_coords in enumerate(tickLevelsWorld):
            for tick_pos in tick_coords[1]:
                base_tick_offsets = self.baseCoordTicks[jdx].copy()
                t_0 = base_tick_offsets[0].copy()
                t_T = base_tick_offsets[1].copy()
                t_0[self.direction] = tick_pos
                t_T[self.direction] = tick_pos
                
                tick_arr.append(t_0)
                tick_arr.append(t_T)
                
                if self.showGrid:
                    g_0 = self.baseCoordGrid[0].copy()
                    g_T = self.baseCoordGrid[1].copy()
                    g_0[self.direction] = tick_pos
                    g_T[self.direction] = tick_pos
                    tick_arr.append(g_0)
                    tick_arr.append(g_T)
                    
        if self._border_flag:
            tick_arr.append(self.border_cells[0]) 
            tick_arr.append(self.border_cells[1]) 
        

        tick_arr = np.array(tick_arr)   
        self.setData(pos=tick_arr)




class GLAxisTextItem(GL3DTextItem):
  
    def __init__(self,
                 tickLevel: float=None,
                 local_position: np.ndarray=None,
                 ref: int=None,
                 parentItem=None, 
                 **kwargs
                 ):        
        super().__init__(parentItem=parentItem, **kwargs)

        self._idRef=ref
        self._tickLevel=tickLevel
        self.local_position=local_position
        
    def setTickLevel(self, tickLevel):
        self._tickLevel=tickLevel

    def idRef(self):
        return self._idRef

    @QtCore.Slot(object)
    def updateTextVisibilityForItem(self, axis_item: GLAxisItem):
        if self.idRef() is None:
            return
        if self.idRef() > axis_item._currentTicks[self._tickLevel]-1:
            self.hide()
        else:
            self.show()
            
    def tickLevel(self): return self._tickLevel



@dataclass(kw_only=True)
class ObjectPosition:
    name: str = field(default=None)
    pos: InitVar[np.ndarray] = field(default=None)
    offset: InitVar[Dict[str, float]] = field(default=None)
    parent: InitVar[ObjectPosition|None] = field(default=None)    
    
    _offset: np.ndarray = field(init=False, default=None)
    _pos: np.ndarray = field(init=False, default=None)
    _position: np.ndarray = field(init=False, default=None)
    _localPosition: np.ndarray = field(init=False, default=None)
    _parent: ObjectPosition = field(init=False, default=None)

    _children: List[ObjectPosition] = field(default_factory=list)
    
    def __post_init__(self, pos, offset, parent):
        if pos is None:
            pos = np.zeros(3, dtype=float)
        if offset is None:
            offset = np.zeros(3, dtype=float)

        self.setOffset(offset)
        self.setPos(pos)
        self.setLocalPosition(self._pos + self._offset)
        self._children = []
        
        self.setParent(parent)  
        if self.name == "tickBorder":
            pprint(self)

    def children(self):
        return self._children
    
    def getParent(self):
        return self._parent
    
    def getPosition(self):
        return self._position

    def setPosition(self, value):
        self._position=value
    
    def getLocalPosition(self):
        return self._localPosition
    
    def setLocalPosition(self, value):
        self._localPosition=value

    def getPos(self):
        return self._pos
    
    def setPos(self, pos):
        self._pos = pos
        self.update()
        
    def setOffset(self, offset):
        self._offset=offset
        
    def update(self):
        self.setLocalPosition(self.getPos() + self.getOffset())
        print(self.name)
        print(self.getParent())
        self.setPosition(self.getLocalPosition() + (self.getParent().getPosition() if self.getParent() else 0))
        for child in self._children:
            child.update()
    
    def getOffset(self):
        return self._offset
    
    def setParent(self, parent: 'ObjectPosition'=None):
        if parent == self.parent:
            return 
        
        if self.getParent() is not None:
            self.getParent().children().remove(self)
            parent.children().append(self)
        
        self._parent = parent
        
        
        
        self.update()    
        
        
    def diff(self, offset_container):
        return self.getPosition() - offset_container.getPosition()
        
        
    
    @classmethod
    def fromParent(cls, kwargs, parent=None):
        offset_container = cls(**kwargs)
        offset_container.setParent(parent)    
        return offset_container
    
        


class GLAxisItem(GL3DGraphicsItemMixin):
    sigAxisChanged = QtCore.Signal(QtCore.QObject)
    
    
    """
    
    axis indexes are ordered x == 0, y == 1, z == 2
       
    
    
    """
    
    def __init__(self,
                 direction: int,
                 origin: Tuple[int, int, int],
                 planeCorner: Tuple[int, int, int],
                 orthogonalVertices: Dict[str, Tuple[float, float, float]]=None,
                 
                 planeDirection: int=1,
                 text: str="",
                 labelFont: QtGui.QFont | None=QtGui.QFont("Neue Haas Grotesk", 16),
                 labelColor: str=getConfigOption('foreground'),
                 labelOffsetPlane: float=0.3, 
                 labelOffsetDirection: float=0., 
                 labelResizeToWorld: bool=True,
                 labelAutoUpdateAnchor: bool=None,
                 labelAnchor: Tuple[float, float]=(0.5, 0.5),
                 showLabel: bool=None,
                 valuesFont: QtGui.QFont | None=QtGui.QFont("Neue Haas Grotesk", 10),
                 valuesColor: Tuple[float,...]=getConfigOption('foreground'),
                 tickLength: float=0.07,
                 ValuesOffset: float=0.01,
                 tickColor: Tuple[float,...]=getConfigOption('foreground'),
                 linkView: GL3DViewBox=None,
                 parentItem=None,
                 maxTickLength=-5,
                 showPlaneBorder: bool=True,
                 showValues: bool=True,
                 tickOffset: float=0.,
                 tickBorderOffset: float=0.1,
                 planeOffset: float=1.,
                 invertedFromWorld: bool=False,
                 autoValueAnchor: bool=True,
                 valuesAnchor: Tuple[float, float]=(0., 0.),
                 valuesResizeToWorld: bool=True,
                 valuesAutoUpdateAnchor: bool=True,
                 showGrid: bool=True,
                 **kwargs
                 ):
        super().__init__(parentItem=parentItem)
      
        
        self.setFlag(self.GLGraphicsItemFlag.ItemHasNoContents)
        self.setParentItem(parentItem)

        self._nTicksMajor = 6
        
        self._linkedView = None
        self._name=None
        
        self.style = {
                    'tickTextOffset': [5, 2],  ## (horizontal, vertical) spacing between text and axis
                    'tickTextWidth': 30,  ## space reserved for tick text
                    'tickTextHeight': 18,
                    'autoExpandTextSpace': True,  ## automatically expand text space if needed
                    'autoReduceTextSpace': True,
                #   'hideOverlappingLabels': hide_overlapping_labels,
                    'valuesFont': valuesFont,
                    "valuesColor" : valuesColor,
                    "tickColor" : tickColor,
                    'stopAxisAtTick': (False, False),  ## whether axis is drawn to edge of box or to last tick
                    'textFillLimits': [  ## how much of the axis to fill up with tick text, maximally.
                        (0, 0.8),    ## never fill more than 80% of the axis
                        (2, 0.6),    ## If we already have 2 ticks with text, fill no more than 60% of the axis
                        (4, 0.4),    ## If we already have 4 ticks with text, fill no more than 40% of the axis
                        (6, 0.2),    ## If we already have 6 ticks with text, fill no more than 20% of the axis
                                    ],
                    'showValues': showValues,
                    "showPlaneBorder"  : showPlaneBorder,
                    "tickLength" : tickLength,
                    "maxNticksperLevel" : 10,
                    "tickWidth" : [2.5, 1.5],
                    'maxTickLevel': 1,
                    'maxTextLevel': 1,
                    'tickAlpha': None,  ## If not none, use this alpha for all ticks.
                    "invertedFromWorld" : int(invertedFromWorld),
                    "valuesAnchor" : valuesAnchor,
                    "autoValueAnchor" : autoValueAnchor,
                    "showGrid" : showGrid,
                }
        self.textWidth = 30  ## Keeps track of maximum width / height of tick text
        self.textHeight = 18
        self._currentTicks=[0 for _ in range(self.style["maxTickLevel"] + 1)]
        
        self.offsetStyle = {"planeOffset" : planeOffset,
                            "tickBorder" : tickBorderOffset,
                            "ticks" : tickOffset,
                            "tickValues" : ValuesOffset,
                            "labelOffsetPlane" : labelOffsetPlane,
                            "labelOffsetDirection" : labelOffsetDirection,
                            }
        
        
        self.ObjectPositionContainer = {
                                }
        
        self.offsetVectors =  {"planeOffset" : None,
                               "tickBorder" : None,
                               "ticks" : None,
                               "tickValues" : None,
                               "labelOffsetPlane" : None,
                               "labelOffsetDirection" : None,
                               }
        
        self._baseGeometry = {"direction" : direction,
                              "orthogonalVertices" : orthogonalVertices,
                              "origin" : np.array(origin),
                              "planeCorner" : np.array(planeCorner),
                              "gridPlaneCoordinates" : None,
                              "grid_f" : None,
                              "borderCoordinates" : None,
                              "planeOrientation" : [None, None],
                             "ticks" : [],
                             "tickValues" : None,     
                             "planeDirection" : None,
                           }
    
        self.fixedWidth = None
        self.fixedHeight = None

        self.logMode = False

        self._tickDensity = 1.0   # used to adjust scale the number of automatically generated ticks
        self._tickLevels  = None  # used to override the automatic ticking system with explicit ticks
        self._tickSpacing = None  # used to override default tickSpacing method
        self.scale = 1.0
        self.autoSIPrefix = True
        self.autoSIPrefixScale = 1.0

        self.labelText = ""
        self.labelUnits = ""
        self.labelUnitPrefix = ""
        self._labelItem = None
        self._GLItems = {"values" : None,
                         "ticks" : None, 
                         "tickLineset" : None,
                         "label" : None
                         }
        self._GLItems: Dict[str, GL3DTextItem|GL3DLinePlotDataItem]
        
        self.tickStyle = {"valuesResizeToWorld" : valuesResizeToWorld,
                          "valuesAnchor" : valuesAnchor,
                          "valuesAutoUpdateAnchor" : valuesAutoUpdateAnchor or valuesAnchor != (0, 0),
                          "showValues" : showValues,
                          'valuesFont': valuesFont,
                          "tickLength" : tickLength,
                          
                          }
        self._siPrefixEnableRanges = None
        
        
        showLabel = False
        
        
        showLabel = showLabel if not bool(text) is None else showValues
        labelAutoUpdateAnchor = showLabel and (labelAutoUpdateAnchor or labelAnchor != (0, 0))
        self.labelStyle = {"labelText" : text,
                           "labelColor" : labelColor or getConfigOption('foreground'),
                           "labelFont" : labelFont,
                           "labelAnchor" : labelAnchor,
                           "labelAutoUpdateAnchor" : labelAutoUpdateAnchor,
                           "labelResizeToWorld" : labelResizeToWorld,
                           "labelOffsetPlane" : labelOffsetPlane,
                           "labelOffsetDirection" : labelOffsetDirection,
                           "showLabel" : showLabel,
                           }
        
        self._baseTickWorldCoords = None
        
        self._buildBaseCoords()
        
        self._GLItems["planeBorder"] = self._buildPlaneBorderCoordinates()
        self._GLItems["values"] = self._initValueTextItems()
        self._GLItems["label"] = self._initLabelTextItem()
        self._GLItems["tickLineset"]= self._initTickItems()
        
        self.setRange(self._baseGeometry["origin"][0], self._baseGeometry["orthogonalVertices"]["direction"][0])
        self.setLabel(**self.labelStyle)

        self._linkedView = None
        
        if linkView is not None:
            self._linkToView_internal(linkView)
            
        self.setTickStyle(valuesColor=valuesColor,
                          valuesFont=QtGui.QFont("Neue Haas Grotesk", 10),
                          valuesAnchor=self.style["valuesAnchor"],
                          )
    def _setViewBoxRange(self, viewBoxRange):
        self.viewBoxRange=viewBoxRange
    def _setDataHelper(self):...
        
    def setData(self):...
    def clipDataFromVRange(self):...

    def _buildPlaneBorderCoordinates(self) -> opengl.GLLinePlotItem:
        p1 = self._baseGeometry["origin"].copy()
        p2 = self._baseGeometry["origin"].copy()
        p2[self.axisDirection()] = self._baseGeometry["planeCorner"][self.axisDirection()]
        
        p4 = self._baseGeometry["origin"].copy()
        p4[self._baseGeometry["planeDirection"]] = self._baseGeometry["planeCorner"][self._baseGeometry["planeDirection"]]
        p3 = self._baseGeometry["planeCorner"]
        
        segments = sum([[a, b] for a, b in zip([p1, p2, p3, p4], [p2, p3, p4, p1])], [])
        
        borderline = opengl.GLLinePlotItem(pos=segments,
                                            color= self.style["tickColor"],
                                            glOptions="translucent",
                                            mode="lines",
                                            )
        borderline.setParentItem(self)
        return borderline

    def _validate_kwargs(self, direction, orthoCoords, perpendicular_axis, viewBoxRange):

        c_viewBoxRange = viewBoxRange.copy() if isinstance(viewBoxRange, list) else viewBoxRange.tolist().copy()
        _ = c_viewBoxRange.pop(direction)
        orthoCoords = list(orthoCoords)
        if not all(orthoCoords[idx] in ax for idx, ax in enumerate(c_viewBoxRange)):
            raise ValueError("Orthogonal Coordinates do not match the viewBox range")

    
    @classmethod
    def fromViewBoxRange(cls,
                         viewRange: list[list[float]] | Dict[str, list[float]],   # limits of the viewBox in world coordinates [[x_min, x_max], [y_min, y_max], [z_min, z_max]] or {"x" : [x_min, x_max], "y" : [y_min, y_max], "z" : [z_min, z_max]}
                         direction: str | int,
                         extremums: Tuple[int, int, int]|Dict[str, int]=None, #    extremum values must be either 0 or 1, and orients the axis origin in world coordinates. {"x" : int, "y" : int ,..} or (int, int ,int)
                         planeDirection: str | int | None = None,         # planeDirection is the direction that shares the rectangular-plane with the axis-direction along the viewRange. e.g. for the unit-cube with direction==x planeDirection==y for the xy plane
                         **kwargs
                         ) -> 'GLAxisItem':
        
        if planeDirection is None or extremums is None:
            raise ValueError(f"If axisVertexPoints is None, planeDirection and extremums needs to be provided")
        else:
            planeDirection = validationHelpers.verifyPlaneDirection(direction, planeDirection)
            extremums = validationHelpers.verifyExtrumums(extremums)
            perpendicularDirection = np.delete([0, 1, 2], [direction, planeDirection]).item()
            
            axisOrigin = [None, None, None]
            
            for idx, ext in enumerate(extremums):
                axisOrigin[idx] = viewRange[idx][ext]

            directionVertex = axisOrigin.copy()
            extremum_idx = extremums[direction]
            
            directionVertex[direction] = viewRange[direction][1-extremum_idx]

            planeVertex = axisOrigin.copy()
            
            orthogonal_extremum_idx = 1 - extremums[planeDirection]
            
            planeVertex[planeDirection] = viewRange[planeDirection][orthogonal_extremum_idx]
            
            perpendicularVertex = axisOrigin.copy()
            orthogonal_extremum_idx = 1 - extremums[perpendicularDirection]
            perpendicularVertex[perpendicularDirection] = viewRange[perpendicularDirection][orthogonal_extremum_idx]

            orthogonalVertices = {"direction" : directionVertex,
                                  "planeDirection" : planeVertex,
                            #      "perpendicularDirection" : perpendicularVertex
                                  }
            
            planeDimensions = [0, 0, 0]
            
            planeDimensions[direction] = directionVertex[direction] - axisOrigin[direction]
            planeDimensions[planeDirection] = planeVertex[planeDirection] - axisOrigin[planeDirection]
            
            offset_mag = 0.15
            
            
            if "planeOffset" in kwargs:
                planeOffset = 3*[0]
                planeOffset[perpendicularDirection] = -1 * offset_mag * viewRange[perpendicularDirection][orthogonal_extremum_idx] / abs(viewRange[perpendicularDirection][orthogonal_extremum_idx])
                kwargs["planeOffset"] = planeOffset

            axisItem  = cls(direction=direction,
                            origin=axisOrigin,
                            planeCorner=planeDimensions,
                            orthogonalVertices=orthogonalVertices,
                            planeDirection=planeDirection,
                            **kwargs
                            )    

            return axisItem

    def _childPaint(self):...
    def _paintHelper(self):... 

    def _buildLabelBaseCoordinates(self):        
        midPoint = np.zeros(3)
        
        midPoint_ax_dir = 0.5 * (self._baseGeometry["planeCorner"][self.axisDirection()] + self._baseGeometry["origin"][self.axisDirection()])
        midPoint[self.axisDirection()] = midPoint_ax_dir
        #midPoint[self.axisDirection()] = 0.5 * (self._baseGeometry["origin"][self.axisDirection()] + orthogonalVertices["direction"][self.axisDirection()])
        
        offset = -1*self._baseGeometry["planeOrientation"][self._baseGeometry["planeDirection"]] * self.offsetStyle["labelOffsetPlane"]
        offset_arr = np.zeros(3)
        offset_arr[self._baseGeometry["planeDirection"]]=offset
        kwargs = {"name":"label",
                  "pos":midPoint,
                  "offset" : offset_arr
                  }
        
        self.ObjectPositionContainer["label"] = ObjectPosition.fromParent(kwargs,
                                                                            parent=self.ObjectPositionContainer["tickEnd"][0])        

        self._baseGeometry["label"] = self.ObjectPositionContainer["label"].getPosition()
        
        
    def _buildGridBaseCoordinates(self):
        gridStart = self._baseGeometry["origin"].copy()        
        kwargs = {"name":"planeOffset",
                  "pos" : gridStart,
                  "offset": np.array(self.offsetStyle["planeOffset"])
                  }
        
        self.ObjectPositionContainer["planeoffset_start"] = ObjectPosition.fromParent(kwargs)
        
        gridStart = self.ObjectPositionContainer["planeoffset_start"].getPosition()
        
        gridEnd_local_position = np.zeros(3)
        gridEnd_local_position[self._baseGeometry["planeDirection"]] = self._baseGeometry["planeCorner"][self._baseGeometry["planeDirection"]] - self._baseGeometry["origin"][self._baseGeometry["planeDirection"]]
        kwargs = {"name":"planeoffset_end",
                  "pos" : gridEnd_local_position,
                  }
        
        self.ObjectPositionContainer["planeoffset_end"] = ObjectPosition.fromParent(kwargs,
                                                                              parent=self.ObjectPositionContainer["planeoffset_start"])

        gridEnd = self.ObjectPositionContainer["planeoffset_end"].getPosition()
        self._baseGeometry["gridPlaneCoordinates"] = (gridStart.tolist().copy(), gridEnd.tolist().copy())
        
        


    def _buildtickBorderCoordinates(self):
        position = np.zeros(3)
        position[self._baseGeometry["planeDirection"]] = -1 * self._baseGeometry["planeOrientation"][self._baseGeometry["planeDirection"]] * self.offsetStyle["tickBorder"]
        kwargs = {"name":"tickBorder",
                  "pos":position,
                    }

        offsetContainer =  ObjectPosition.fromParent(kwargs,
                                                       parent=self.ObjectPositionContainer["planeoffset_start"]
                                                       )

        self.ObjectPositionContainer["tickBorder"] = offsetContainer
        
        
        borderPositionStart = self.ObjectPositionContainer["tickBorder"].getPosition()
        #borderPositionStart[self.axisDirection()] = self._baseGeometry["origin"][self.axisDirection()]
        
        pos = np.zeros(3)
        pos[self.axisDirection()] = self._baseGeometry["planeCorner"][self.axisDirection()]
        kwargs = {"name":"tickBorder",
                   "pos":pos}
        
        endPositionContainer = ObjectPosition.fromParent(kwargs, self.ObjectPositionContainer["tickBorder"])
        borderPositionEnd = endPositionContainer.getPosition()
        
        borderPositon = np.vstack((borderPositionStart, borderPositionEnd))

        
        self._baseGeometry["borderCoordinates"] = [borderPositon[0,:].tolist(), borderPositon[1,:].tolist()]
        
      
        
        
    def _builderTickBaseCoord(self):
        self._baseGeometry["tickStart"]=[]
        self._baseGeometry["tickEnd"]=[]
        self._baseGeometry["tickValues"]=[]


        orthogonalVertices = self._baseGeometry["orthogonalVertices"]
        tickOrientationToPlane = self._baseGeometry["origin"][self._baseGeometry["planeDirection"]] - orthogonalVertices["planeDirection"][self._baseGeometry["planeDirection"]]

        self.ObjectPositionContainer["tickStart"] = [None for idx in range(self.style["maxTickLevel"] + 1)]
        self.ObjectPositionContainer["tickEnd"] = [None for idx in range(self.style["maxTickLevel"] + 1)]

        self.ObjectPositionContainer["tickValues"] = [None for idx in range(self.style["maxTickLevel"] + 1)]
        
        for idx in range(self.style["maxTickLevel"] + 1):
            
            tickLocalPositionStart = np.zeros(3)
            offset = np.array([tickOrientationToPlane * self.offsetStyle["ticks"] if idx == self._baseGeometry["planeDirection"] else 0 for idx in range(3)])
            
            kwargs = {"name":"tickStart",
                        "pos":tickLocalPositionStart,
                        "offset":offset
                        }
            
            line_length_mag = self.tickStyle["tickLength"] / (1 + idx)
            
            offsetContainerTick =  ObjectPosition.fromParent(kwargs,
                                                               parent=self.ObjectPositionContainer["tickBorder"]
                                                        )
            
            tickEndPos = np.zeros(3)
            tickEndPos[self._baseGeometry["planeDirection"]] = -1 * self._baseGeometry["planeOrientation"][self._baseGeometry["planeDirection"]] * line_length_mag
            
            OffsetTickEnd = ObjectPosition.fromParent({"name":None,     
                                                         "pos" :  tickEndPos,                                             
                                                         },
                                                        parent=offsetContainerTick)
            

            
            offset = np.array([tickOrientationToPlane * self.offsetStyle["tickValues"] if idx == self._baseGeometry["planeDirection"] else 0 for idx in range(3)])
            kwargs = {"name":"tickValues",
                      "offset": offset
                      }
            
            offsetContainerTickValues =  ObjectPosition.fromParent(kwargs,
                                                                    parent=OffsetTickEnd
                                                                    )

            self.ObjectPositionContainer["tickStart"][idx] = offsetContainerTick
            self.ObjectPositionContainer["tickEnd"][idx] = OffsetTickEnd
            self.ObjectPositionContainer["tickValues"][idx] = offsetContainerTickValues
            
            tickStart = offsetContainerTick.getPosition()
            tickEnd = OffsetTickEnd.getPosition()
          #  tickEnd[self._baseGeometry["planeDirection"]] = tickStart[self._baseGeometry["planeDirection"]] + -1 *tick_length_with_dir
            
            tick_coords = [tickStart, tickEnd]
            print(f'tick_coords: {tick_coords}')
            
            self._baseGeometry["ticks"].append(tick_coords)
            
            tickValuesStart = offsetContainerTickValues.getPosition()
            
            self._baseGeometry["tickValues"].append(tickValuesStart)



    def _buildBaseCoords(self):
        plane_diff = self._baseGeometry["planeCorner"] - self._baseGeometry["origin"]
        
        self._baseGeometry["planeOrientation"] = [diff / abs(diff) if diff != 0 else 0 for diff in plane_diff]
        self._baseGeometry["planeDirection"] = [direction for direction, orientation in enumerate(self._baseGeometry["planeOrientation"]) if direction!=self.axisDirection() and orientation != 0][0]
        
        self._buildGridBaseCoordinates()
        
        
        self._buildtickBorderCoordinates()
        self._builderTickBaseCoord()
        self._buildLabelBaseCoordinates()
                
        
    
    def axisDirection(self):
        return self._baseGeometry["direction"]
                
    
    def tickValues(self, minVal:float, maxVal:float, size: float):
        minVal, maxVal = sorted((minVal, maxVal))

        minVal *= self.scale
        maxVal *= self.scale

        ticks = []
        ticksWorld = []
        
        
        tickLevels = self.tickSpacing(minVal, maxVal, size)
        allValues = np.array([])
        for i in range(len(tickLevels)):
            spacing, offset = tickLevels[i]

            ## determine starting tick
            start = (math.ceil((minVal-offset) / spacing) * spacing) + offset

            ## determine number of ticks
            num = int((maxVal-start) / spacing) + 1
            values = (np.arange(num) * spacing + start) / self.scale
            ## remove any ticks that were present in higher levels
            ## we assume here that if the difference between a tick value and a previously seen tick value
            ## is less than spacing/100, then they are 'equal' and we can ignore the new tick.
            close = np.any(
                np.isclose(
                    allValues,
                    values[:, np.newaxis],
                    rtol=0,
                    atol=spacing/self.scale*0.01
                ),
                axis=-1
            )
            values = values[~close]
            allValues = np.concatenate([allValues, values])
            ticks.append((spacing/self.scale, values.tolist()))
            ticksWorld.append((self.getWorldCoord(spacing/self.scale), self.getWorldCoord(values).tolist()))
            self._currentTicks[i] = len(values.tolist())
        return ticks, ticksWorld

    def updateTickValues(self, tickLevels, tickLevelsWorld):
        
        for jdx, (ticks, ticksWorld) in enumerate(zip(tickLevels, tickLevelsWorld)):
            for idx, (tick, tickWorld) in enumerate(zip(ticks[1], ticksWorld[1])):
                text_item=self._GLItems["values"][jdx][idx]
                text_item.local_position[self.axisDirection()] = tickWorld
                text_item.blockUpdates(True)
                text_item.setData(pos=text_item.local_position, text=str(round(tick,2)))
                text_item.blockUpdates(False)
                

  
    def setRange(self, mn: float, mx: float):
        self.range = [mn, mx]
        tickLevels, tickLevelsWorld = self.tickValues(*self.range, 1)
        if self.style["showValues"]:
            self.updateTickValues(tickLevels, tickLevelsWorld)
        self._GLItems["tickLineset"].updateTickPositions(tickLevelsWorld)
        self.sigAxisChanged.emit(self)        
        self.update()
        
        
    def setTextFont(self, font=None):
        if font is None:
            getConfigOption("foreground")
            
    def setLabel(
                self,
                labelText: str | None=None,
                labelColor: str | None=None,
                labelFont: str | None=None,
                labelAnchor: str=None,
                labelResizeToWorld: bool=None,
                labelAutoUpdateAnchor: bool=None,
                showLabel: bool=None,
                **kwargs):  
        
        
        if not labelText is None:
            
            
            self.labelStyle["labelText"] = labelText
        
        visible = bool(self.labelStyle["labelText"])
        if visible:
            self._GLItems["label"].setData(text=self.labelStyle["labelText"])
        else:
            self._GLItems["label"].setData(text="")
        if not showLabel is None:
            if showLabel and not visible:
                raise
            
        self.showLabel(visible)
        
        del kwargs["labelOffsetPlane"]
        del kwargs["labelOffsetDirection"]

        if not labelColor is None:
            self.labelStyle["labelColor"] = labelColor
        if not labelFont is None:
            self.labelStyle["labelFont"] = labelFont
        if not labelAnchor is None:
            self.labelStyle["labelAnchor"] = labelAnchor
        if not labelResizeToWorld is None:
            self.labelStyle["labelResizeToWorld"] = labelResizeToWorld and visible
            self.labelStyle["labelText"]
        
        if not labelAutoUpdateAnchor is None:
            self.labelStyle["labelAutoUpdateAnchor"] = (labelAutoUpdateAnchor or labelAnchor != (0, 0)) and visible
        
        kwargs = {#"resizeToWorld" : self.labelStyle["labelResizeToWorld"],
                  "text" : self.labelStyle["labelText"],
                  "color" : self.labelStyle["labelColor"],
                  "font" : self.labelStyle["labelFont"],
             #     "anchor" : self.labelStyle["labelAnchor"],
                  }
        
        

        self._GLItems["label"].setData(**kwargs)
    
    
    
    def orthogonalVertices(self) -> Dict[str, Tuple[float, float, float]]:
        return self._baseGeometry["orthogonalVertices"]
        
        
    def showLabel(self, show: bool=True):
        if show:
            self._GLItems["label"].show()
        else:
            self._GLItems["label"].hide()
    
    def setTickStyle(self,
                     tickColor: str | None=None,
                     tickAnchor: Tuple[float, float]=(0., 0.),
                     valuesFont: str | None=None,
                     valuesResizeToWorld: bool=True,
                     valuesAutoUpdateAnchor: bool=None,
                     showValues: bool=None,
                     *args,
                     **kwargs
                     ):
        
        if not tickColor is None:
            self.tickStyle["tickColor"] = tickColor            
        if not valuesFont is None:
            self.tickStyle["valuesFont"] = valuesFont            
        if not tickAnchor is None:
            self.tickStyle["tickAnchor"] = tickAnchor
        if not valuesResizeToWorld is None:
            self.tickStyle["valuesResizeToWorld"] = valuesResizeToWorld
        if not tickAnchor is None:
            self.tickStyle["tickAnchor"] = tickAnchor
        if not showValues is None:
            self.tickStyle["showValues"] = showValues
        
        if not valuesAutoUpdateAnchor is None:
            self.tickStyle["valuesAutoUpdateAnchor"] = (valuesAutoUpdateAnchor or tickAnchor != (0, 0)) and self.tickStyle["showValues"]

        if self.tickStyle["showValues"]:
        
            if "values" in self._GLItems and valuesFont:
                for tick_level in self._GLItems["values"]:
                    for item in tick_level:
                        item.setData(font=self.tickStyle["valuesFont"])
    
    def _initLabelTextItem(self) -> GLAxisTextItem:
        labeItem = GLAxisTextItem(local_position=self._baseGeometry["label"],
                                  pos=self._baseGeometry["label"],
                                  anchor=self.labelStyle["labelAnchor"],
                                  font=self.labelStyle["labelFont"], 
                                  )
        labeItem.setParentItem(self)
        return labeItem
        
    
    def _initValueTextItems(self) -> List[List[GLAxisTextItem]]:
        gl_objects_level = []
        idx=0
        for jdx in range(self.style["maxTickLevel"] + 1):
            gl_objects = []
            for idx in range(self.style["maxNticksperLevel"]):
                local_position = self._baseGeometry["tickValues"][jdx].copy()
                gl_object = GLAxisTextItem(tickLevel=jdx,
                                           local_position=local_position,
                                           ref=idx,
                                           anchor=self.tickStyle["valuesAnchor"],
                                           font=self.tickStyle["valuesFont"], 
                                           )
                gl_object.setParentItem(self)
                gl_object.hide()
                self.sigAxisChanged.connect(gl_object.updateTextVisibilityForItem)
                gl_objects.append(gl_object)
                idx+=1
                
            gl_objects_level.append(gl_objects)
        return gl_objects_level

    def _initTickItems(self) -> _TickLineset:
        tickItem = _TickLineset(self.axisDirection(),
                               baseCoordTicks=self._baseGeometry["ticks"],
                               baseCoordGrid=self._baseGeometry["gridPlaneCoordinates"],
                               borderCoordinates=self._baseGeometry["borderCoordinates"],
                               glOptions="translucent",
                               antialias=True,
                               color=self.style["tickColor"],
                               width=self.style["tickWidth"][0],
                               showGrid=self.style["showGrid"],
                               showBorder=self.style["showPlaneBorder"],
                               )
        tickItem.setParentItem(self)
        return tickItem
        
    def checkWidths(self, textNum: str):...
    
    @QtCore.Slot(object)
    @QtCore.Slot(object, object)
    def linkedViewChanged(self, view: GL3DViewBox, newRange=None):
   #     print(newRange)
        if newRange is None:
            newRange = view.viewRange()[self.axisDirection()]
        self.setRange(*newRange)

    def tickSpacing(self, minVal: float, maxVal: float, size: float):
        _tickDensity=1
        dif = abs(maxVal - minVal)
        if dif == 0:
            return []

        ref_size = 300. # axes longer than this display more than the minimum number of major ticks
        minNumberOfIntervals = max(
            2.25,       # 2.0 ensures two tick marks. Fudged increase to 2.25 allows room for tick labels.
            2.25 * _tickDensity * math.sqrt(size/ref_size) # sub-linear growth of tick spacing with size
        )

        majorMaxSpacing = dif / minNumberOfIntervals

        mantissa, exp2 = math.frexp(majorMaxSpacing)
        p10unit = 10. ** (
            math.floor(
                (exp2-1)
                / 3.32192809488736
            ) - 1
        )
        
        if 100. * p10unit <= majorMaxSpacing:
            majorScaleFactor = 10
            p10unit *= 10.
        else:
            for majorScaleFactor in (50, 20, 10):
                if majorScaleFactor * p10unit <= majorMaxSpacing:
                    break
        
        majorInterval = majorScaleFactor * p10unit
        levels = [
            (majorInterval, 0),
        ]

        if self.style['maxTickLevel'] >= 1:
            # Modified this section to choose a smaller interval for minor ticks
            if majorScaleFactor == 10:
                minorScaleFactor = 5  # Half of major interval for minor ticks
            elif majorScaleFactor == 20:
                minorScaleFactor = 10  # Half of major interval
            elif majorScaleFactor == 50:
                minorScaleFactor = 25  # Half of major interval
            else:
                minorScaleFactor = majorScaleFactor / 2  # Default to half
            
            minorInterval = minorScaleFactor * p10unit
            levels.append((minorInterval, 0))

        if self.style['maxTickLevel'] >= 2:
            # Keep the extra level logic but ensure it's smaller than minor interval
            if majorScaleFactor == 10:
                extraScaleFactor = 2  # Smaller than minor interval
            elif majorScaleFactor == 20:
                extraScaleFactor = 5  # Smaller than minor interval
            elif majorScaleFactor == 50:
                extraScaleFactor = 10  # Smaller than minor interval
            else:
                extraScaleFactor = minorScaleFactor / 2  # Default to half
            
            extraInterval = extraScaleFactor * p10unit
            if extraInterval < minorInterval:
                levels.append((extraInterval, 0))
        return levels

    def getWorldCoord(self, value):
        view = self.linkedView()
        if view is None:
            world_bounds = 0, 1
        else:
            world_bounds = view.state["worldRange"][self.axisDirection()]

        dif = self.range[1] - self.range[0]
        rmin, rmax = self.range
        if rmax == rmin:
            scale = 1
            offset = 0
            value
        else:
            scale = (world_bounds[1] - world_bounds[0]) / dif
            offset = world_bounds[0] - rmin * scale
        return scale * value + offset
    
    def labelString(self):
        return self.labelStyle["labelText"]
        
    def _linkToView_internal(self, view: GL3DViewBox):
        self.unlinkFromView()
        self._linkedView = weakref.ref(view)

        self._setViewBoxRange(view.state["worldRange"])
        
        if self.axisDirection() == 0:
            view.sigXRangeChanged.connect(self.linkedViewChanged)
        elif self.axisDirection() == 1:
            view.sigYRangeChanged.connect(self.linkedViewChanged)
        else:
            view.sigZRangeChanged.connect(self.linkedViewChanged)
        view.sigResized.connect(self.linkedViewChanged)
        if isinstance(view.parentWidget(), GLViewWidget):
            plotitem_widget = view.parentWidget()
                        
            if self.tickStyle["valuesAutoUpdateAnchor"] or self.labelStyle["labelAutoUpdateAnchor"]:
                    plotitem_widget.sigViewAngleChanged.connect(self.autoUpateAnchor)
                    self.autoUpateAnchor(plotitem_widget)


    @classmethod
    def compute_azimuth(cls, origin, point):
        dx = point[0] - origin[0]
        dy = point[1] - origin[1] 
        angle = math.atan2(dy, dx)
        return math.degrees(angle) % 360

    @classmethod
    def compute_dp(cls, origin, point):
        dx = point[0] - origin[0]
        dy = point[1] - origin[1]
        perp_slope = -dx/dy if dy != 0 else float('inf')

        m = perp_slope
        
        length = 0.2
        if np.isinf(m):
            dx, dy = 0, length / 2
        else:
            dx = (length / 2) / np.sqrt(1 + m**2)
            dy = m * dx
        return perp_slope, dx, dy
    
    @classmethod
    def clockwise_angle(cls, p1, p2, k2):
        v1 = p2 - p1
        v2 = k2 - p1

        dot = np.dot(v1, v2)
        det = v1[0]*v2[1] - v1[1]*v2[0]  

        angle_rad = np.arctan2(det, dot) 
        clockwise_angle = (-angle_rad) % (2 * np.pi)
        return np.degrees(clockwise_angle)
        
    
    @QtCore.Slot(object)
    def autoUpateAnchor(self, plotitem_widget: GL3DViewWidget):
        
        angle = GLAxisItem.clockwise_angle(self._baseGeometry["origin"][:2],
                                           self.orthogonalVertices()["direction"][:2],
                                           plotitem_widget.cameraPositionNumpy()[:2]
                                           )
#        print(f"\n{self.axisDirection()}")
#        print(f"angle: {angle}")
        half = int(angle // 90) % 4
        

        if half % 4==0:
            newAnchor = (1, self.style["valuesAnchor"][1])
        else:
            newAnchor = (0, self.style["valuesAnchor"][1])
        
        if self.labelStyle["labelAutoUpdateAnchor"]:
            self._GLItems["label"].setAnchor(newAnchor)
            
        if self.tickStyle["valuesAutoUpdateAnchor"]:
            if newAnchor != self.style["valuesAnchor"]:
                self.style["valuesAnchor"]=newAnchor
                for text_item_container in self._GLItems["values"]:
                    for text_item in text_item_container:
                        text_item.setAnchor(newAnchor)
                        
        
    def linkToView(self, view):
        self._linkToView_internal(view)

    def unlinkFromView(self):
        oldView = self.linkedView()
        self._linkedView = None
        if oldView is not None:
            oldView.sigResized.disconnect(self.linkedViewChanged)
            if self.axisDirection() == 0:
                oldView.sigXRangeChanged.disconnect(self.linkedViewChanged)
            elif self.axisDirection() == 1:
                oldView.sigYRangeChanged.disconnect(self.linkedViewChanged)
            else:
                oldView.sigZRangeChanged.disconnect(self.linkedViewChanged)

    def linkedView(self) -> GL3DViewBox:
        return None if self._linkedView is None else self._linkedView()



    def setTickFont(self, font: QtGui.QFont | None):
        self.style['tickFont'] = font
        self.picture = None
        #self.prepareGeometryChange()
        self.update()






from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, Any
if TYPE_CHECKING:
    from ...instruments.utils import InstrumentManager
    from ...quantities.engines import TickEngineManager
    from ...instruments.instruments import Option
    from PySide6.QtCore import Signal 
    from custom_widgets import CustomPlotDataItem
    from ...data_classes.classes import Slice

from PySide6.QtWidgets import (QMainWindow, QWidget, QDockWidget, QSplitter, 
                              QVBoxLayout, QLabel, QApplication, QWidget)

from PySide6.QtCore import Qt
from PySide6 import QtCore
from . import axis_widgets  
from .. import subplot
import pyqtgraph as pg
from .. import custom_widgets
import time

class SplitDockWidget(QDockWidget):
    __AXIS_DEPENDENCE__ ="xy"

    
    def __init__(self,
                 *,
                 tick_engine_manager: TickEngineManager=None,
                 **kwargs
                 ):
        super().__init__(**kwargs)   
        print(f"\n\n\nSplitDockWidget")
        print(f"tick_engine_manager.get_engine: {tick_engine_manager.get_engine}")
        print(f"\n\n\n")
        self._dclass_name_subplot_title_map: Dict[str, str]={"skew" : "Skew",
                                                             "term" : "Term Structure"
                                                             }
        
        self._initLayout()
        self._initSubplots(self._dclass_name_subplot_title_map.values(), tick_engine_manager)
        
        
    
    def _initSubplots(self,
                      subplot_names: List[str],
                      tick_engine_manager: TickEngineManager=None,
                      ):
        
        x_ax_2D = axis_widgets.CustomAxisItem(axis_direction=0,
                                              tick_engine=tick_engine_manager.get_engine("x") if tick_engine_manager is not None else None,
                                              orientation="bottom"
                                              )
        y_ax_2D = axis_widgets.CustomAxisItem(axis_direction=1,
                                              tick_engine=tick_engine_manager.get_engine("y") if tick_engine_manager is not None else None,
                                              orientation="bottom")
        z_ax_2D = axis_widgets.CustomAxisItem( axis_direction=2,
                                              tick_engine=tick_engine_manager.get_engine("z") if tick_engine_manager is not None else None,
                                              orientation="right") 
        zz_ax_2D = axis_widgets.CustomAxisItem(axis_direction=2,
                                               tick_engine=tick_engine_manager.get_engine("z") if tick_engine_manager is not None else None,
                                               orientation="right") 
        
        axes_2D_items = [x_ax_2D, y_ax_2D, z_ax_2D, zz_ax_2D]
        
        
        splitter: QSplitter = self.widget().layout().itemAt(0).widget()
        
        for i, (title, axis) in enumerate(zip(subplot_names, SplitDockWidget.__AXIS_DEPENDENCE__)):
            tick_engine = tick_engine_manager.get_engine(SplitDockWidget.__AXIS_DEPENDENCE__[-i-1])
            axisItems={"bottom" : axes_2D_items[i],
                       "right"   : axes_2D_items[i+2]
                       }
            subplot_item = SubPlotItem(title=title,
                                        axis_3D_dir=axis,
                                        other_axis_tick_engine=tick_engine,
                                        axisItems=axisItems
                                        )
            subplot_widget = pg.PlotWidget(plotItem=subplot_item)
            splitter.addWidget(subplot_widget)

    def _initLayout(self):
        container = QWidget()
        layout = QVBoxLayout(container)
        splitter = QSplitter(Qt.Vertical)
        layout.addWidget(splitter)
        container.setLayout(layout)
        self.setWidget(container)



class SubPlotItem(pg.PlotItem):
    def __init__(self,
                 axis_3D_dir: str,
                 other_axis_tick_engine=None,
                 *args,
                 **kwargs
                 ):
        super().__init__(*args, **kwargs)
        self._axisSlice = 0 if axis_3D_dir=="x" else 1
        self.interacting=False
        self.setAcceptHoverEvents(True)
        self.axis_3D_directions=axis_3D_dir
        self.other_axis_tick_engine=other_axis_tick_engine
        self._other_axis_direction = "x" if axis_3D_dir == "yz" else "y"
        self._plotDataItemContainer: Dict[str, CustomPlotDataItem]={}
        self._dataclassContainer: Dict[str, Slice]={}
        self.show_text=False
        self.prev_x=None
        self.prev_y=None
        self.interacting=False
        self.other_axis_textitem = custom_widgets.CustomTextItem(text="",
                                                                 view_box_anchor=(1, 0),
                                                                 color=(0, 0, 0, 255),
                                                                 fill=pg.mkBrush(255, 255, 255, 200),
                                                                 anchor=(1, 0)
                                                                 ) 
        self.other_axis_textitem.attach_viewbox(self.getViewBox())
        self.other_axis_textitem.hide()
        self.addItem(self.other_axis_textitem)
        
        self.getViewBox().setAutoVisible(x=False, y=False)
        self.getViewBox().setBackgroundColor("k")
    
    def axisSlice(self):
            return self._axisSlice
    
        
    def hoverLeaveEvent(self, event):
        self.interacting=False
        super().hoverLeaveEvent(event)

    def sceneEventFilter(self, watched, event):
        if event.type() == QtCore.QEvent.GraphicsSceneMousePress\
            or event.type() == QtCore.QEvent.GraphicsSceneMouseMove\
            or event.type() == QtCore.QEvent.GraphicsSceneMouseRelease\
            or event.type() ==  QtCore.QEvent.GraphicsSceneWheel:
                
            self.interacting=True
        return super().sceneEventFilter(watched, event)
        
    def addItem(self, item):
        if isinstance(item, custom_widgets.CustomPlotDataItem):
            self._plotDataItemContainer[item.px_type] = item
        super().addItem(item)
        
    def removeItem(self, item):
        if isinstance(item, custom_widgets.CustomPlotDataItem):
            del self._plotDataItemContainer[item.px_type]
        super().removeItem(item)
        del item
        
    def setParentItem(self, parent):
        super().setParentItem(parent)
        if not parent is None:
            if self.width() > 0:
                self.other_axis_textitem.setPos(self.width(), 0)

    def mousePressEvent(self, ev):
        self.interacting=True
        return super().mousePressEvent(ev)
    
    def mouseReleaseEvent(self, ev):
        self.interacting=False
        return super().mouseReleaseEvent(ev)
    
    def listPricePlotDataItem(self):
        return list(self._plotDataItemContainer.values())
    
    def pricePlotDataItem(self, px_type):
        return self._plotDataItemContainer[px_type]
    
    def set_text(self, x, y):
        values = locals()[self._other_axis_direction]
        text = self.other_axis_tick_engine.function([values])[0]
        text = f"{self.other_axis_tick_engine.get_label()}: {text}"
        self.other_axis_textitem.setPlainText(text)
    
    def addDataClass(self,
                     name: str,
                     dataclass_item: Slice
                     ):
        self._dataclassContainer[name]=dataclass_item
    
    def removeDataClass(self, name: str):
        del self._dataclassContainer[name]
    
    def dataClass(self, name):
        return self._dataclassContainer.get(name, None)    
    
    def plotFromDataClass(self, dataclass: Slice, **kwargs) -> CustomPlotDataItem:
        if dataclass.valid_values:
            kwargs.update(dataclass.plot_item_kwargs())
        else:
            kwargs.update({"x" : [], "y" : []})

        plot_data_item = custom_widgets.CustomPlotDataItem(**kwargs)
        dataclass.add_update_callback(plot_data_item.update_from_dataclass)
        
        self.addDataClass(plot_data_item.px_type, dataclass)
        self.addItem(plot_data_item)
        return plot_data_item
    
    def removePricePlotDataItem(self, px_type: str) -> None:
        plot_data_item = self.pricePlotDataItem(px_type)
        dataclass = self.dataClass(px_type)
        dataclass.remove_update_callback(plot_data_item.update_from_dataclass)
        
        self.removeDataClass(px_type)
        self.removeItem(plot_data_item)
        
    @QtCore.Slot(float, float, float)
    def update3DMouseClick(self, x, y, z):
        for dataclass in self._dataclassContainer.values():
            dataclass.update_point(x, y, z)
    
    
    



from PySide6 import QtCore, QtGui, QtWidgets    
app = QtWidgets.QApplication([])


def close():
    parent.close()
    import time
    time.sleep(2)
    import sys
    sys.exit()




parent = QtWidgets.QWidget()
parent.setWindowTitle("Parent Widget")
parent.resize(300, 200)
layout = QtWidgets.QVBoxLayout(parent)
label_parent = QtWidgets.QLabel("This is the parent widget")
layout.addWidget(label_parent)


timer = QtCore.QTimer()
timer.timeout.connect(close)
timer.start(2000)


child = QtWidgets.QWidget(parent)
child.setWindowTitle("Child Widget")
child.resize(150, 100)
child.move(50, 50)
label_child = QtWidgets.QLabel("This is the child widget", child)
child.show()



parent.show()
app.exec()



import sys
import numpy as np
from PySide6 import QtCore, QtGui, QtWidgets    
from pyqtgraph import opengl
from OpenGL import GL, GLU
from pprint import pprint   

def create_grid(rotation, translation):
    grid = opengl.GLGridItem()
    grid.setSize(x=1, y=1, z=1)
    grid.setSpacing(x=0.2, y=0.2, z=0.2)
    grid.rotate(*rotation)
    grid.translate(*translation)
    return grid

def create_txt(pos):
    return opengl.GLTextItem(text = f"{pos}", pos=pos, font=QtGui.QFont("Arial", 9))
    
    

def map_2D_coords_to_3D(widget, x, y):
    widget_width = widget.width()
    widget_height = widget.height()
    device_pixel_ratio = widget.window().screen().devicePixelRatio()

    ndc_x = x / widget_width
    ndc_y = y / widget_height

    viewport = GL.glGetIntegerv(GL.GL_VIEWPORT)
    
    _, _, viewport_width, viewport_height = viewport

    mouse_x_physical = ndc_x * viewport_width
    mouse_y_physical = ndc_y * viewport_height
    mouse_y_physical = viewport_height - mouse_y_physical 
    
    
    modelview = np.array(widget.viewMatrix().data()).reshape(4, 4)
    projection = np.array(widget.projectionMatrix(viewport, (0, 0, device_pixel_ratio * widget_width, device_pixel_ratio* widget_height)).data()).reshape(4, 4)
    print("proj:")
    pprint(projection.round(4).tolist())
    print("view:")
    pprint(modelview.round(4).tolist())
    depth = GL.glReadPixels(int(mouse_x_physical), int(mouse_y_physical), 1, 1, GL.GL_DEPTH_COMPONENT, GL.GL_FLOAT)[0][0]
    print(f"depth: {depth}")
    world_x, world_y, world_z = GLU.gluUnProject(mouse_x_physical, mouse_y_physical, depth, modelview, projection, viewport)

    return world_x, world_y, world_z

def map_3D_coords_to_2D(widget, world_x, world_y, world_z):
    device_pixel_ratio = widget.window().screen().devicePixelRatio()
    widget_height = widget.height()
    widget_width = widget.width()
    
    viewport = GL.glGetIntegerv(GL.GL_VIEWPORT)
    _, _, viewport_width, viewport_height = viewport
    
    modelview = np.array(widget.viewMatrix().data()).reshape(4, 4)
    projection = np.array(widget.projectionMatrix(viewport,
                                                (0, 0, device_pixel_ratio * widget_width, device_pixel_ratio * widget_height)).data()).reshape(4, 4)
    
    px_x, px_y, px_z = GLU.gluProject(world_x, world_y, world_z, modelview, projection, viewport)
    px_y = viewport_height - px_y
    return px_x, px_y, px_z




def world_to_viewport_pixel(widget: opengl.GLViewWidget, world_coords):
    if len(world_coords) == 3:
        world_coords = np.append(world_coords, 1.0)

    mvp = widget.currentProjection() * widget.currentModelView()
    mvp_np = np.array(mvp.data()).reshape(4, 4)

    clip_coords = mvp_np @ world_coords

    ndc_coords = clip_coords[:3] / clip_coords[3]

    viewport = widget.geometry()
    x_pixel = (ndc_coords[0] + 1) * 0.5 * viewport.width() + viewport.x()
    y_pixel = (1 - (ndc_coords[1] + 1) * 0.5) * viewport.height() + viewport.y()
    z_pixel = (ndc_coords[2] + 1) / 2
    return x_pixel, y_pixel, z_pixel


class CustomGLViewWidget(opengl.GLViewWidget):
    def __init__(self):
        super().__init__()
    
    def mouseMoveEvent(self, ev):
        return super().mouseMoveEvent(ev)
    
    def wheelEvent(self, ev):
        super().wheelEvent(ev)
        self.compute_optimal_clipping()
        return 
    
    def initializeGL(self):
        super().initializeGL()
        GL.glEnable(GL.GL_DEPTH_TEST)  
        return 

    def mousePressEvent(self, ev):
        if ev.buttons() == QtCore.Qt.MouseButton.LeftButton:
            print("\n")
            x_px, y_px = ev.x(), ev.y()
            world = map_2D_coords_to_3D(self, x_px, y_px)
            print(f"window coords: {np.round((x_px, y_px), 3)}")
            print(f"world coords: {np.round(world, 3)}")
            print(f"back to px1: {np.round(world_to_viewport_pixel(self, world),3)}")
            print(f"back to px2: {np.round(map_3D_coords_to_2D(self, *world),3)}")
            _, _, width, height = self.getViewport()
            print(f"width/2: {np.round(width/2,2)}")
            print(f"height/2: {np.round(height/2,2)}")
            
            
        return super().mousePressEvent(ev)
    
    def compute_optimal_clipping(self):
        return 
        largest_point_distance = np.sqrt(3)
        near_clip = max(0.0001, self.opts['distance'] - 1.01 * largest_point_distance) 
        far_clip = self.opts['distance'] 
        
        self.opts["near"] = near_clip
        self.opts["far"] = far_clip
        self.update()

    
    
  #  def keyPressEvent(self, ev):
       # if ev.key() == QtCore.Qt.Key.Key_T:
            
      #  return super().keyPressEvent(ev)
#
def create_pos():
    X = np.arange(-5, 5, 0.5)
    Y = np.arange(-5, 5, 0.5)
    X_mat, Y_mat = np.meshgrid(X, Y, indexing="xy")
    R = np.sqrt(X_mat**2 + Y_mat**2)
    

    m1 = 3
    
    Z_mat = m1 * np.sin(R) 
    X = (X - X.min()) / (X.max() - X.min())
    Y = (Y - Y.min()) / (Y.max() - Y.min())
    Z_mat = (Z_mat - Z_mat.min()) / (Z_mat.max() - Z_mat.min())

    
    return X, Y, Z_mat

class Window(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()
        
        self.gl_widget = CustomGLViewWidget()
        self.setCentralWidget(self.gl_widget)
        self.gl_widget.opts["distance"] = 5
        
        self.g1 = create_grid(rotation=(0, 0, 0, 0), translation=(0.5, 0.5, 0))
        self.g2 = create_grid(rotation=(90, 0, 1, 0), translation=(0, 0.5, 0.5))
        self.g3 = create_grid(rotation=(90, 1, 0, 0), translation=(0.5, 0, 0.5))
        
        text_items = [create_txt((0,0,0)), create_txt((0,0,1)),create_txt((1,0,0)),create_txt((0,1,0))] 
        for txt_item in text_items:
            self.gl_widget.addItem(txt_item)


        self.gl_widget.addItem(self.g1)
        self.gl_widget.addItem(self.g2)
        self.gl_widget.addItem(self.g3)
        
        X, Y, Z_mat = create_pos()
        Z_mat += 0.5
        self.surface = opengl.GLSurfacePlotItem(X, Y, Z_mat, glOptions='opaque', color = (1, 0, 1, 1))
        
        self.gl_widget.addItem(self.surface)
        
        self.showMaximized()
        
def main():
    app = QtWidgets.QApplication()
    win = Window()
    app.exec()


if __name__ == "__main__":
    main()


from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, Any
if TYPE_CHECKING:
    from engines.yield_engines.implied_engines import AbstractYieldEngine 

from abc import ABC, abstractmethod
from pyvol_terminal.engines.option_engines import interpolation
import numpy as np
import math
import time
from datetime import datetime
from .. import utils
from dataclasses import dataclass, field, fields
from . import quantities
from ..engines.option_engines.base import ABCEuroEngine



@dataclass(slots=True, frozen=True, kw_only=True)
class _BaseSpecifications:
    ticker: str
    px_quotation: float=None

    
    def __repr__(self):
        cls_name = self.__class__.__name__
        non_none = {f.name: getattr(self, f.name) for f in fields(self) if getattr(self, f.name) is not None}
        args_str = ", ".join(f"{k}={v!r}" for k, v in non_none.items())
        return f"{cls_name}({args_str})"


@dataclass(slots=True, kw_only=True)
class Quotes:
    last_ts: float=np.nan
    px: Dict[str, float]
    size: Dict[str, float]
    
    open_interest: float=np.nan
    
    callbacks: List[Callable]=field(default=list)

    
    def update(self,
               timestamp: float,
               px_bid=None,
               px_ask=None,
               size_bid=None,
               size_ask=None,
               open_interest=None,
                )-> None:
        self.last_ts=timestamp
        
        self.px["bid"]=px_bid
        self.px["ask"]=px_ask
        self.px["mid"]=0.5 * (px_bid + px_ask)

        self.size["bid"]=size_bid
        self.size["ask"]=size_ask

        if not open_interest is None:
            self.open_interest = open_interest
        
    def copy(self):
        return list(self, self.px.keys())
        
    def add_callback(self, callback):
        self.callbacks.append(callback)
    
    def remove_quote_callback(self, callback):
        self.callbacks.remove(callback)
    


class ABCPyVol(ABC):
    
    def update(self, timestamp, px_bid=None, px_ask=None, callbacks=True) -> None: 
        self.update_internal(timestamp, px_bid, px_ask, callbacks)
    
    @abstractmethod
    def update_internal(self): ...


class ABCInstrument(ABCPyVol):
    def __init__(self,
                 specs: _BaseSpecifications,
                 *args, 
                 **kwargs
                 ) -> None:
        super().__init__(*args, **kwargs)
        self.specs=specs
        self.quotes=Quotes()

        self._metric_state_container=[]
        self._child_derivative_callbacks=[]
        self._metric_callbacks=[]
        self._structure_callbacks=[]
        

    def update_quotes(self, *args, **kwargs) -> None:
        self.quotes.update(*args, **kwargs)
        for callback in self.quotes.callbacks:
            callback(self)


    def update_metrics(self):...
                        
    def update_internal(self, *args, **kwargs) -> None:
        self.update_quotes(*args, **kwargs)   
            
        self.update_metrics()
        
        for callback in self._child_derivative_callbacks:
            callback()
        for callback in self._structure_callbacks:
            callback(self)
    
    def add_quote_callback(self, callback):
        self.quotes.add_callback(callback)
    
    def remove_quote_callback(self, callback):
        self.quotes.add_callback(callback)

    def add_child_derivative_callback(self, callback):
        self._child_derivative_callbacks.append(callback)
    
    def remove_child_derivative_callback(self, callback):
        self._child_derivative_callbacks.remove(callback)
    
    def add_structure_callback(self, callback):
        self._structure_callbacks.append(callback)

    def remove_structure_callback(self, callback):
        self._structure_callbacks.remove(callback)

    def add_price_callbacks(self, callback):
        self._price_callbacks.append(callback)
        
    def remove_price_callbacks(self, callback):
        self._price_callbacks.remove(callback)

    def add_metric_callbacks(self, callback):
        self._metric_callbacks.append(callback)
        
    def remove_metric_callbacks(self, callback):
        self._metric_callbacks.remove(callback)
        
    def modify_metric_state(self, metric, state, rebuild=True):
        if state=="add":
            self._metric_state_container.append(metric)
        else:
            self._metric_state_container.remove(metric)
        
        if rebuild:
            self._build_metric_state_container()
    
    def _build_metric_state_container(self):
        ...


@dataclass(slots=True, kw_only=True, frozen=True)
class SpotSpecifications(_BaseSpecifications):
    category: str=field(default="spot", init=False)
  
class Spot(ABCInstrument):
    def __init__(self, specs, *args, **kwargs):
        self.specs=specs
        super().__init__(*args, **kwargs)
    
    def update_metrics(self): ...   


@dataclass(slots=True, kw_only=True, frozen=True)
class DerivativeSpecifications(_BaseSpecifications):
    specs_underlying: _BaseSpecifications=None
    settlement_px_quotation: str=None
    
    def __post_init__(self):
        if self.settlement_px_quotation is None:
            self.settlement_px_quotation=self.px_quotation


class Derivative(ABCInstrument):
    def __init__(self,
                 *args, 
                 **kwargs
                 ):
        self.quotes_underlying: Quotes=None
        super().__init__(*args, **kwargs)

    @abstractmethod
    def update_metrics(self): ...   
    
    @abstractmethod
    def underlying_listener(self): ...
    

@dataclass(slots=True, kw_only=True, frozen=True)
class FutureSpecifications(DerivativeSpecifications):
    expiry: float=None
    category: str=field(default="future", init=False) 

    
class Future(Derivative):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.specs: FutureSpecifications
    
    def update_from_underlying(self, underlying_object):
        self._underlying_pxs=underlying_object.copy()
    
    def get_forward_rate(self):
        yte = utils.convert_unix_maturity_to_years(self.specs.expiry)         
        self.forward_rate = np.log(self / getattr(self.underlying_px, self.underlying_px_type)) / yte
        
    def update_metrics(self):...
    

@dataclass(slots=True, frozen=True, kw_only=True)
class OptionSpecifications(Derivative):
    category: str=field(default="option", init=False)
    expiry: float
    strike: float
    option_type: str
    exercise_style: str
    delivery: str=None
    
    
    def __post_init__(self):
        if self.option_type is not None and self.option_type not in ["c", "p"]:
            raise
        elif self.exercise_style is not None and self.exercise_style not in ["eu", "am"]:
            raise
        elif self.delivery is not None and self.exercise_style not in ["p", "c"]:
            raise


class Option(Derivative):
    def __init__(self,
                 *args,
                 ivol_engine: Any,
                 itm_conversion_engine=None, 
                 **kwargs
                 ) -> None:
        
        super().__init__(*args, **kwargs)
        self.specs: OptionSpecifications
        self.ivol_engine: ABCEuroEngine=ivol_engine

        self.states={px_type : quantities.OptionState({"flag" : self.specs.option_type,
                                                       "strike" : self.specs.option_type
                                                       }) for px_type in ["bid", "ask", "mid"]
                     }
        self.ivol_metrics: List[quantities.IVOLMetrics]
        self.itm_conversion_engine = itm_conversion_engine if itm_conversion_engine is not None else interpolation.ITMConversionEngine()
        
        self.listening_px_types = {"option" : [],
                                   "underlying" : []
                                   }
        
        self.underlying_target_px: List[str]=[]
        
        self._cached_underlying_quotes: Dict[str, float] = {"bid" : np.nan,
                                                            "ask" : np.nan,
                                                            "mid" : np.nan
                                                            }
        self._underlying_px_targets=set()
        
        self._ptype_idx_map = {"px_bid" : 0,
                               "px_ask" : 1,
                               "px_mid" : 2
                               }
    
    def refresh_quotes(self, *args) -> None:
        self._refresh("option")
    
    def refresh_quotes_underlying(self, *args) -> None:
        self._refresh("underlying")
        
    def _refresh(self, instrument: str) -> None:
        """
        
        Skip if instrument being refresh has cache. This method is only called when we get response from instrument, therefore
        another metric was what changed. 
        
        """
        for ivol_metrics in self.ivol_metrics:
            ivol_config = ivol_metrics.config
            px_option = self.quotes[ivol_config.px_type_option]
            px_underlying = self.quotes_underlying.px[ivol_config.px_type_underlying]            
            ts_px = self.quotes.last_ts, px_option if instrument == "option" else self.quotes_underlying.last_ts, px_underlying
            
            if math.isnan(ts_px[1]) or ivol_metrics.is_cached(*{instrument : ts_px}):
                continue
            else:
                if not self.states[ivol_config.px_type_underlying].check_set_get("OTM", px_underlying):
                    continue
                
                timestamp = max(self.quotes.last_ts, self.quotes_underlying.last_ts)
                ivol = self.calc_ivol(timestamp, px_option, px_underlying)
                                                       
                ivol_metrics.update(ivol,
                                    {"option" : [self.quotes.last_ts, px_option],
                                     "underlying" : [self.quotes_underlying.last_ts, px_underlying],
                                     })
            metric_names = ivol_metrics.get_req_metrics()
            metric_values = self.ivol_engine.calculate(metric_names)
            ivol_metrics.set_metrics(**metric_values)
            
            
    def calc_ivol(self, ts, px_opt, px_under):
        """
        
        extra call for people to easily override if subclassing the Option class.

        Parameters
        ----------
        ts : float
            epoch timestamp in (s)
        px_opt : float
            _description_
        px_under : float
            _description_

        Returns
        -------
        _type_
            _description_
        """
        return self.ivol_engine.calculate_IVOL(ts, px_opt, px_under)

            
    def update_ivol(self) -> None:
        for ivol_metrics in self.ivol_metrics:
            ivol_config = ivol_metrics.config
            
            quotes_underlying = self.quotes_underlying

            if quotes_underlying:
                pass
            
            
            ivol = self.ivol_engine.calculate_IVOL(max(self.quotes.last_ts, quotes_underlying.last_ts),
                                                   getattr(self.quotes, ivol_config.px_type_option),
                                                   getattr(self.quotes_underlying, ivol_config.px_type_underlying),
                                                   )
            ivol_metrics.update(ivol)
            
        
    def create_ivol_metric(self, config) -> None:
        ivol_config = quantities.IVOLConfig(**config)
        
        if all(len(self.ivol_metrics) > 0,
               not math.isnan(self.quotes[ivol_config.px_type_option]),
               not math.isnan(self.quotes_underlying[ivol_config.px_type_underlying])
            
            ##### We have to check if other variables are also non-nan #####
            
            ):
            ts = np.nan
        else:
            ts=np.nan
        

        ivol_metrics = quantities.IVOLMetrics(config=ivol_config,
                                              last_ts=ts,
                                              greeks=quantities.Greeks(),
                                              moneyness=quantities.Moneyness()
                                              )
        self.ivol_metrics.append(ivol_metrics)
        self._underlying_px_targets.update(ivol_config.px_type_underlying)
        
    
    def add_active_ivol_ptype(self, opt_pytpe) -> None:
        if opt_pytpe not in self.active_ivol_ptypes:
            self.active_ivol_ptypes = self.active_ivol_ptypes + (opt_pytpe,)
        
    def remove_active_ivol_ptype(self, opt_ptype) -> None:
        self.active_ivol_ptypes = tuple(ivol for ivol in self.active_ivol_ptypes if ivol != opt_ptype)
    
    def update_metrics(self, engine=None, cachedIV=False):
        self.calculate_implied_volatility(engine, cachedIV)
        self.calculate_greeks(engine)
        self.calculate_moneyness_metrics(engine)
        
    def update_from_OTM_sister(self, sister_option: "Option", engine=None):
        ivol_b, ivol_a = self.itm_conversion_engine.ITM_ivol(sister_option, self)
        self.ivol[self._ptype_idx_map["px_bid"]] = ivol_b
        self.ivol[self._ptype_idx_map["px_ask"]] = ivol_a
        self.ivol[self._ptype_idx_map["px_mid"]] = sister_option.ivol[self._ptype_idx_map["px_mid"]]
        self.calculate_greeks(engine)
        self.calculate_moneyness_metrics()
    
    def get_all_metrics_price_type(self, px_type):
        idx = self._ptype_idx_map[px_type]
        self.underlying_px = self.underlying_px
        return self.prices[idx], self.ivol[idx], self.delta[idx], self.delta_mag[idx], self.gamma[idx], self.vega[idx], self.theta[idx], self.rho[idx], self.moneyness, \
                self.log_moneyness, self.standardised_moneyness[idx], self.OTM, self.call_flag, self.underlying_px

    def get_all_metrics(self):# -> tuple[Any, Any, Any, Any, Any, Moneyness | Any | float, f...:
        return self.ivol, self.delta, self.delta_mag, self.gamma, self.vega, self.moneyness, self.log_moneyness,\
                self.standardised_moneyness, self.OTM, self.underlying_px
        
            
    def calculate_implied_volatility(self, option_engine=None, cachedIV=False):
        if cachedIV:
            return
        if not math.isnan(self.underlying_px) and self.OTM:
            if option_engine is None:
                option_engine = self.ivol_engine
            
            for px_type in self.active_ivol_ptypes:
                self.ivol[self._ptype_idx_map[px_type]] = option_engine.calculate_IVOL(self.last_update_time,
                                                                                    getattr(self, px_type),
                                                                                    self.underlying_px
                                                                                    )
                print(self.ivol)
        else:
            self.ivol=self._nan_3_numpy.copy()

    def calculate_greeks(self, option_engine=None):
        if not math.isnan(self.underlying_px):
            if option_engine is None:
                option_engine = self.ivol_engine
            self.delta, self.gamma, self.vega, self.theta, self.rho = option_engine.calculate_all_greeks(self.last_update_time,
                                                                                                        self.ivol,
                                                                                                        self.underlying_px)          
            self.delta_mag = np.abs(self.delta)
        else:
            self.delta, self.gamma, self.vega, self.theta, self.rho, self.delta_mag = [self._nan_3_numpy.copy()] * 6

    def get_underlying(self, *args):
        return self.underlying_px
    
    def get_ivol(self, px_type=None):
        if px_type is None:
            return self.ivol
        return self.ivol[self._ptype_idx_map[px_type]]
    
    def get_delta(self, px_type=None):
        if px_type is None:
            return self.delta
        return self.delta[self._ptype_idx_map[px_type]]

    def get_gamma(self, px_type=None):
        if px_type is None:
            return self.gamma
        return self.gamma[self._ptype_idx_map[px_type]]

    def get_vega(self, px_type=None):
        if px_type is None:
            return self.vega
        return self.vega[self._ptype_idx_map[px_type]]

    def get_theta(self, px_type=None):
        if px_type is None:
            return self.theta
        return self.theta[self._ptype_idx_map[px_type]]

    def get_rho(self, px_type=None):
        if px_type is None:
            return self.rho
        return self.rho[self._ptype_idx_map[px_type]]
    
    def get_greeks(self, px_type=None):
        if px_type is None:
            return self.delta, self.gamma, self.vega, self.theta, self.rho
        idx = self._ptype_idx_map[px_type]
        return self.delta[idx], self.gamma[idx], self.vega[idx], self.theta[idx], self.rho[idx]
    
    def moneyness_metrics(self, px_type=None):
        if px_type is None:
            return self.moneyness, self.log_moneyness, self.standardised_moneyness
        idx = self._ptype_idx_map[px_type]
        return self.moneyness, self.log_moneyness, self.forward_moneyness, self.standardised_moneyness[idx], 
    
    def calculate_moneyness_metrics(self, *arg):
        if not math.isnan(self.underlying_px):
            self.moneyness = self.specs.strike / self.underlying_px
            if not self.future_underyling:
                forward_factor = 0
                if not self.ivol_engine.interest_rate_engine is None:
                    forward_factor += self.ivol_engine.interest_rate_engine.evaluate(utils.convert_unix_maturity_to_years(self.specs.expiry))
                if not self.ivol_engine.dividend_rate_engine is None:
                    forward_factor += self.ivol_engine.dividend_rate_engine.evaluate(utils.convert_unix_maturity_to_years(self.specs.expiry))

                forward_rate = math.exp(forward_factor * utils.convert_unix_maturity_to_years(self.specs.expiry))
                self.forward_moneyness = forward_rate * self.moneyness
            
            self.log_moneyness = math.log(self.moneyness)
            
            if self.future_underyling:
                self.standardised_moneyness = self.log_moneyness / (self.ivol * math.sqrt(utils.convert_unix_maturity_to_years(self.specs.expiry)))
            else:
                self.standardised_moneyness = math.log(self.forward_moneyness) / (self.ivol * math.sqrt(utils.convert_unix_maturity_to_years(self.specs.expiry)))
        else:
            self.moneyness=np.nan
            self.log_moneyness=np.nan
            self.standardised_moneyness = self._nan_3_numpy.copy()

  
@dataclass(slots=True, frozen=True, kw_only=True)
class OptionInvSpecifications(OptionSpecifications):
    category: str=field(default="opi", init=False)


class OptionInverted(Option):
    """
    inverted options are essentially FX options settle in foreign ccy
    """
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
    
    
    
    
    def calc_ivol(self, ts, px_opt, px_under):
        
        return super().calc_ivol(ts, px_under*px_opt, px_under)
    


class ExchangeTradedProduct(ABCInstrument):
    def __init__(self, tickers, bpx_asket_instrument_type, **kwargs):
        super().__init__(tickers, **kwargs)
        self.object_bpx_asket=[]
        for ticker in tickers:
            self.object_bpx_asket.append(getattr(bpx_asket_instrument_type(ticker)))
            
    def update_from_underlying(self, underlying_object):
        self.underlying_px=getattr(underlying_object, self.underlying_px_type)
    
    def update_metrics(self):...




def get_metric_attr_all():
    metric_dict = { "price" : {"multi_dim_flag": True},
                    "strike" : {"multi_dim_flag": False},
                    "expiry" : {"multi_dim_flag": False},
                    "ivol": {"multi_dim_flag": True},
                    "delta": {"multi_dim_flag": True},
                    "gamma": {"multi_dim_flag": True},
                    "vega": {"multi_dim_flag": True},
                    "theta": {"multi_dim_flag": True},
                    "rho": {"multi_dim_flag": True},
                    "moneyness": {"multi_dim_flag": False},
                    "log_moneyness": {"multi_dim_flag": False},
                    "standardised_moneyness" : {"multi_dim_flag": True},
                    "forward_moneyness" : {"mult_dim_flag" : False},
                    "OTM": {"multi_dim_flag": False},
                    "call_flag": {"multi_dim_flag": False},
                    "underlying_px": {"multi_dim_flag": False},
                }
    return metric_dict

def get_metric_category(metric):
    metric_dict = {
        "gamma": "greek",
        "delta": "greek",
        "vega": "greek",
        "theta": "greek",
        "rho": "greek",
        "mid": "price",
        "bid": "price",
        "ask": "price",
        "ivol": "volatility",
        "TVAR" : "volatility",
        "VAR" : "volatility", 
        "IVM" : "volatility",
        "IVA" : "volatility",
        "IVB" : "volatility",
        "strike": "strike",
        "expiry": "expiry",
        "moneyness": "moneyness",
        "log_moneyness": "moneyness",
        "standardised_moneyness": "moneyness",
        "forward_moneyness": "moneyness",
        "OTM": "flag",
        "call_flag": "flag",
        "underlying_px": "price"
    }
    
    return metric_dict.get(metric, None)


def get_metric_attr():
    metric_dict = { "price" : {"multi_dim_flag": True},
                    "strike" : {"multi_dim_flag": False},
                    "expiry" : {"multi_dim_flag": False},
                    "ivol": {"multi_dim_flag": True},
                    "OTM": {"multi_dim_flag": False},
                    "call_flag": {"multi_dim_flag": False},
                    "underlying_px": {"multi_dim_flag": False},
                }
    return metric_dict

def greeks_attr():
    metric_dict = {"delta": {"multi_dim_flag": True},
                    "gamma": {"multi_dim_flag": True},
                    "vega": {"multi_dim_flag": True},
                    "theta": {"multi_dim_flag": True},
                    "rho": {"multi_dim_flag": True}
                    }
    return metric_dict

def moneyness_attr():
    metric_dict = {"moneyness": {"multi_dim_flag": False},
                   "log_moneyness": {"multi_dim_flag": False},
                   "standardised_moneyness": {"multi_dim_flag": True},
                   "forward_moneyness" : {"mult_dim_flag" : False},
                    }
    return metric_dict


import sys
import numpy as np
from PySide6 import QtCore, QtGui
from PySide6.QtWidgets import QApplication
import pyqtgraph.opengl as gl
from pyvol_terminal.interfaces.volatility_surface.normalised_glviewwidget import NormaliseGLViewWidget
from pyvol_terminal.interfaces.volatility_surface import view_box
from pyvol_terminal.interfaces.volatility_surface.axis_3D_items import AxisItem3D
from pyvol_terminal.interfaces.volatility_surface.gl_plotitems import glScatter, glSurface
from pyvol_terminal import misc_widgets
class Counter:
    c=0


def main():
    app = QApplication(sys.argv)

    padding = {ax : 0 for ax in "xyz"}
    w = NormaliseGLViewWidget(padding=padding)
    labels = "XYZ"
    
    axis_items = {axis : AxisItem3D(axis, label, 6, "red", font=QtGui.QFont('Arial', 10)) for axis, label in zip("xyz", labels)}
    w.setAxisItems(axis_items)
    w.setWindowTitle("NormaliseGLViewWidget Test")
    w.opts['azimuth'] = -50
    w.opts["distance"] = 4
    w.opts['center'] = QtGui.QVector3D(0, 1, 0)  


    x = 5 * np.linspace(-2, 2, 25)
    y = 5 * np.linspace(-2, 2, 25)
    x_grid, y_grid = np.meshgrid(x, y)
    r = np.sqrt(x_grid ** 2 + y_grid ** 2) + 1e-6
    z_grid = 3 * np.sin(r) / r
    
    color=(1, 0, 0, 1)
    colormap=misc_widgets.CustomColorMap("virdis")
    
    surface_item = glSurface(x=x, y=y, z=z_grid, price_type="mid", color=color, colormap=colormap)
    surface_item.item_type="surface"
    w.addItem(surface_item, viewBoxAdd=True)
    
    print(f"\nsurface")
    print(f"x: {(surface_item._x.min(), surface_item._x.max())}")
    print(f"y: {(surface_item._y.min(), surface_item._y.max())}")
    print(f"z: {(surface_item._z.min(), surface_item._z.max())}")
    
    
    
    
    scatter_points = np.column_stack((x_grid.flatten(), y_grid.flatten(), z_grid.flatten()))

    scatter_item = glScatter(pos=scatter_points, price_type="mid", color=color)
    scatter_item.item_type="scatter"
    w.addItem(scatter_item, viewBoxAdd=True)

    
    #x_min = np.min((scatter_item.unnorm_pos[:,0].min(), surface_item.unnorm_x.min()))
    #x_max = np.max((scatter_item.unnorm_pos[:,0].max(), surface_item.unnorm_x.max()))

        
    
    for i in range(1, 11):
        pos=(i /10, 0, 0)
        text_item = gl.GLTextItem(pos=pos, text=f"{i+1}", color="white")
        text_item.type="axis"
        w.addItem(text_item, viewBoxAdd=False)
    for i in range(1,11):
        pos=(0, i/10, 0)
        text_item = gl.GLTextItem(pos=pos, text=f"{i+1}", color="yellow")
        text_item.type="axis"
        w.addItem(text_item, viewBoxAdd=False)
    for i in range(1,11):
        pos=(0, 0, i/10)
        text_item = gl.GLTextItem(pos=pos, text=f"{i+1}", color="cyan")
        text_item.type="axis"
        w.addItem(text_item, viewBoxAdd=False)  
    

    counter = Counter()
    
    
    def update_data():
        Counter.c+=1
        print("")
        x = surface_item.unnorm_x + Counter.c
        y = surface_item.unnorm_y + Counter.c
        z = surface_item.unnorm_z + Counter.c
        print("\nupdate_data")
        surface_item.setData(x=x, y=y, z=z)
        pos=scatter_item.unnorm_pos + Counter.c
        scatter_item.setData(pos=pos)
        
        #update_text("x", surface_item)
        #update_text("y", surface_item)
        #update_text("z", surface_item)
#        if Counter.c > 3:
#            sys.exit()
        
        
        
        
    timer = QtCore.QTimer()
    timer.timeout.connect(update_data)
    timer.start(2000)
    
    w.show()
    w.update()

    # Start the event loop
    sys.exit(app.exec())




if __name__ == "__main__":
    main()



from PySide6 import QtWidgets, QtCore
from pyqtgraph.opengl import GLViewWidget
import numpy as np
from OpenGL import GL, GLU
import sys

def map_2D_coords_to_3D(widget, x, y):
    widget_height = widget.height()
    widget_width = widget.width()
    device_pixel_ratio = widget.window().screen().devicePixelRatio()

    x_norm = x / widget_width
    y_norm = y / widget_height

    viewport = GL.glGetIntegerv(GL.GL_VIEWPORT)
    _, _, viewport_width, viewport_height = viewport

    mouse_x_physical = x_norm * viewport_width
    mouse_y_physical = y_norm * viewport_height
    mouse_y_physical = viewport_height - mouse_y_physical 
    
    depth = GL.glReadPixels(int(mouse_x_physical), int(mouse_y_physical),  1, 1, GL.GL_DEPTH_COMPONENT, GL.GL_FLOAT)[0][0]

    modelview = np.array(widget.viewMatrix().data()).reshape(4, 4)
    projection = np.array(widget.projectionMatrix(viewport, (0, 0, device_pixel_ratio * widget_width, device_pixel_ratio* widget_height)).data()).reshape(4, 4)
    
    world_x, world_y, world_z = GLU.gluUnProject(mouse_x_physical, mouse_y_physical, depth, modelview, projection, viewport)
    px_x, px_y, px_z = GLU.gluProject(world_x, world_y, world_z, modelview, projection, viewport)
    
    px_y = viewport_height - px_y
    
    world_coords = world_x, world_y, world_z
    px_coords = px_x, px_y, px_z
    
    return world_coords, px_coords



class CustomGLViewWidget(GLViewWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setMouseTracking(True)
        self.setAttribute(QtCore.Qt.WA_Hover)
        self.setAttribute(QtCore.Qt.WA_DontCreateNativeAncestors) 

    def mousePressEvent(self, event):
        match event.buttons().value :    
            case 1: 
                self.interacting=True
                pos = event.pos()
                world_coords, px_coords = map_2D_coords_to_3D(self, pos.x(), pos.y())
                print(f"\npos: {np.round((pos.x(), pos.y()), 4)}")
                print(f"world: {np.round((world_coords), 4)}")
                print(f"px_coords: {np.round((px_coords), 4)}")

class MainWindow(QtWidgets.QMainWindow):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setWindowTitle('Volatility Surface')
        self.widget_central = QtWidgets.QWidget()
        self.setCentralWidget(self.widget_central)
        self.layout_main = QtWidgets.QVBoxLayout()
        self.widget_central.setLayout(self.layout_main)
        
        self.view_widget = CustomGLViewWidget()
        self.vw_layout = QtWidgets.QVBoxLayout()
        self.layout_main.addLayout(self.vw_layout)
        self.layout_main.addWidget(self.view_widget)
        self.showMaximized()
        
        

if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    mainWin = MainWindow()
    mainWin.show()
    sys.exit(app.exec())
    



from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, Any
if TYPE_CHECKING:
    from pyvol_terminal.utils import AbstractValidPriceChecker 

from abc import ABC, abstractmethod
import numpy as np
from scipy.interpolate import PchipInterpolator, make_interp_spline
from ... import utils
import math
from pprint import pprint



class AbstractYieldEngine:
    
    @abstractmethod
    def fit(self, *args):
        ...
    
    @abstractmethod
    def evaluate(self, *args):
        ...
    

def implied_q_from_PC_BSM(C,P,r,T,S,K):
    q = math.log((C - P + K*np.exp(-r*T))/S)/-T
    return q

def implied_forward_rate(F_1, F_2, T_1, T_2):
    return math.log(F_2 / F_1) / (T_2 - T_1)

def implied_q_from_futures(F, S, T, r):
    return r - math.log(F / S)/T

def implied_r_from_futures(F, S, T, q):
    return q + math.log(F / S)/T

def get_implied_from_PC_black(C, P, F, K, T):
    return - math.log((C-P) / (F - K)) / T

class ImpliedFromFuture(AbstractYieldEngine):
    def __init__(self, model_type, instrument_manager, bid_key="bid", ask_key="ask", given_rate_engine=None):
        self.bid_key=bid_key
        self.ask_key=ask_key
        self.instrument_manager=instrument_manager
        self.model_type=model_type
        self.given_rate_engine=given_rate_engine
        self.instrument_to_idx_map={}
        self.spot_future_name_map={}
        self.mean_discretisation_steps=20
        self._null_arr = np.array([0] * self.mean_discretisation_steps)
        self.interpolator = lambda x: self._null_arr 
        self._expiry_min = None

        if self.model_type == "black":
            self.calculate_implied_yield = implied_r_from_futures
        elif self.model_type == "black_scholes_merton":
            self.calculate_implied_yield = implied_q_from_PC_BSM    
        
        if given_rate_engine is None:
            self.given_rate_engine=DummyYieldClass()
        
        self._setup()
    
    def _setup(self):        
        self._n_future = len(self.instrument_manager.futures)
        self.expiry_arr = np.nan * np.empty(self._n_future)
        self.bid_arr = np.nan * np.empty(self._n_future)
        self.ask_arr = np.nan * np.empty(self._n_future)
        self.mid_arr = np.nan * np.empty(self._n_future)
        self.imp_rate = np.nan * np.empty(self._n_future)   
        self.spot_arr = np.nan * np.empty(self._n_future)   
        
        for idx, (ticker, instrument_object) in enumerate(self.instrument_manager.futures.items()):
            self.instrument_to_idx_map[ticker] = idx
            spot_name = instrument_object.underlying_object.ticker
            if spot_name in self.spot_future_name_map:
                self.spot_future_name_map[spot_name].append(ticker)
            else:
                self.spot_future_name_map[spot_name] = [ticker]
            self.expiry_arr[idx] = float(instrument_object.expiry)
            if self._expiry_min is None:
                self._expiry_min = instrument_object.expiry
            else:
                if instrument_object.expiry < self._expiry_min:
                    self._expiry_min=instrument_object.expiry
                    
    def update_data(self, websocket_responses):
        for ticker, price_dict in websocket_responses.items():
            self._internal_updater(ticker, price_dict[self.bid_key], price_dict[self.ask_key])

    def _internal_updater(self, ticker, bid_price, ask_price):
        
        asset_object=self.instrument_manager.all_instrument_objects[ticker]
        asset_object.update_price(bid_price, ask_price, callbacks=False)
        if ticker in self.instrument_to_idx_map:
            idx = self.instrument_to_idx_map[ticker]
            yte = utils.convert_unix_maturity_to_years(asset_object.expiry)
            self.imp_rate[idx] = self.calculate_implied_yield(asset_object.mid,
                                                            asset_object.underlying_object.mid,
                                                            yte,
                                                            self.given_rate_engine.evaluate(yte))
        if ticker in self.spot_future_name_map:
            for future_names in self.spot_future_name_map[ticker]:
                future_object=self.instrument_manager.all_instrument_objects[future_names]
                idx = self.instrument_to_idx_map[future_names]
                yte = utils.convert_unix_maturity_to_years(future_object.expiry)
                self.imp_rate[idx] = self.calculate_implied_yield(future_object.mid,
                                                                future_object.underlying_object.mid,
                                                                yte,
                                                                self.given_rate_engine.evaluate(yte))

    def fit(self):
        yte = utils.convert_unix_maturity_to_years(self.expiry_arr)
        x, y = utils.filter_nans_2D(yte, self.imp_rate)
        if x.size > 2:
            if np.all(np.diff(self.imp_rate) >= 0):
                self.interpolator = PchipInterpolator(x, y)
            else:
                self.interpolator = make_interp_spline(x, y, k=2)
        else:
            self.interpolator = lambda x: self._null_arr
        
    def evaluate(self, yte):
        xi = np.linspace(utils.convert_unix_maturity_to_years(self._expiry_min), yte, 20)
        return self.interpolator(xi).sum() / 20
        

class ImpliedFromOption(AbstractYieldEngine):
    def __init__(self, model_type, instrument_manager=None, bid_key="bid", ask_key="ask"):
        if model_type == "black":
            self.calculate_implied_yield = get_implied_from_PC_black
        elif model_type == "black_scholes_merton":
            self.calculate_implied_yield = implied_q_from_PC_BSM

        self.instrument_manager=instrument_manager        
        self.exp_yield = {}
        self.expiry_option_pair_map = {}
        self.expiry_opt_list_map = {}
        self.bid_key=bid_key
        self.ask_key=ask_key
        self._find_mid_strike()
        self.mean_discretisation_steps=20
        self._null_arr = np.array([0] * self.mean_discretisation_steps)
    
    def _find_mid_strike(self,):
        strikes = []
        for option_object in self.instrument_manager.options.values():
            strikes.append(option_object.strike)
        self._median_strike = np.median(strikes)
    
    def _find_closest_option(self, option_name_list):
        def _finding(option_name_list, default_val=None):
            smallest_abs_diff = 1e10
            smallest_diff_instrument_name = None
            for option_name in option_name_list:
                option_object = self.instrument_manager.options[option_name]
                if default_val:
                    abs_diff = abs(option_object.strike - default_val)
                else: 
                    abs_diff = abs(option_object.strike - option_object.underlying_object.mid) 
                if abs_diff < smallest_abs_diff:
                    smallest_abs_diff = abs_diff 
                    smallest_diff_instrument_name = option_name
            return smallest_diff_instrument_name
        smallest_diff_instrument_name = _finding(option_name_list)
        if smallest_diff_instrument_name is None:
            smallest_diff_instrument_name = _finding(option_name_list, self._median_strike)
        return smallest_diff_instrument_name
    
    def _create_constructor(self):
        for opt1_name, opt2_name in self.instrument_manager.options_maps.put_call_map.items():
            option_object1 = self.instrument_manager.options[opt1_name]
            expiry = option_object1.expiry
            
            if expiry in self.expiry_opt_list_map:
                list_for_specific_expiry = self.expiry_opt_list_map[expiry]
                if not opt2_name in list_for_specific_expiry:    # checking if the put or call with same expiry and maturity is already in dict.
                    self.expiry_opt_list_map[expiry].append(opt1_name)
            else:
                self.expiry_opt_list_map[expiry] = [opt1_name]
        
        self.expiry_opt_list_map = {key: self.expiry_opt_list_map[key] for key in sorted(self.expiry_opt_list_map)}
        
        for expiry, option_name_list in self.expiry_opt_list_map.items():
            smallest_diff_instrument_name = self._find_closest_option(option_name_list)
            
            if not smallest_diff_instrument_name is None:
                smallest_option_object = self.instrument_manager.options[smallest_diff_instrument_name]
                
                if smallest_option_object.flag_int == 1:
                    call_object = smallest_option_object
                    put_name = self.instrument_manager.options_maps.put_call_map[call_object.ticker]
                    put_object = self.instrument_manager.options[put_name]
                else:
                    put_object = smallest_option_object
                    call_name = self.instrument_manager.options_maps.put_call_map[put_object.ticker]
                    call_object = self.instrument_manager.options[call_name]

                temp_dict = {"call" : call_object,
                             "put" : put_object}

                self.expiry_option_pair_map[expiry] = temp_dict
        
    def _create_yield_curve(self,):
        yte_arr = []
        yield_arr = []
        print(f"\n_create_yield_curve")
        for expiry, object_dict in self.expiry_option_pair_map.items():
            call_object = object_dict["call"]
            put_object = object_dict["put"]
            yte = utils.convert_unix_maturity_to_years(expiry)
            dividend = get_implied_from_PC_black(call_object.mid,
                                                 put_object.mid,
                                                 call_object.underlying_object.mid,
                                                 call_object.strike,
                                                 yte)
            if not np.isnan(dividend):
                print(np.array((call_object.mid,
                                                 put_object.mid,
                                                 call_object.underlying_object.mid,
                                                 call_object.strike,
                                                 yte)))
                print(f"\ndividend: {dividend}")
                self.exp_yield[expiry] = dividend
                yte_arr.append(yte)
                yield_arr.append(dividend)
                
        return np.array(yte_arr), np.array(yield_arr)   

    def update_data(self, websocket_responses):
        for ticker, price_dict in websocket_responses.items():
            self._internal_updater(ticker, price_dict[self.bid_key], price_dict[self.ask_key])
                    
    def _internal_updater(self, ticker, bid_price, ask_price):
        option_object=self.instrument_manager.all_instrument_objects[ticker]
        option_object.update_price(bid_price, ask_price, callbacks=False)
            
    def fit(self):
        self._create_constructor()
        yte_arr, yield_arr = self._create_yield_curve()        
        if len(self.exp_yield) > 2:
            if np.all(np.diff(yield_arr) >= 0):
                self.interpolator = PchipInterpolator(yte_arr, yield_arr)
            else:
                self.interpolator = make_interp_spline(yte_arr, yield_arr, k=2)
        else:
            self.interpolator = lambda x: self._null_arr 
        xi = np.linspace(0, 1, 20)
        print("\nevaluate")
        print(np.round(self.interpolator(xi), 3))

    def evaluate(self, yte):
        xi = np.linspace(0, yte, 20)
        return self.interpolator(xi).sum() / 20


class DummyYieldClass:
    def __init__(self):
        pass
    
    def evaluate(self, x):
        return 0


from PySide6.QtWidgets import QGraphicsTransform
from PySide6.QtGui import QMatrix4x4
from PySide6.QtCore import Property, QObject
from PySide6 import QtCore, QtGui, QtWidgets
from PySide6.QtCore import QTimer, QPropertyAnimation


class CustomTransform(QGraphicsTransform):
    
    def __init__(self, parent=None):
        super().__init__(parent)
        self._vertical_shear = 0.0 
        self._horizontal_shear = 0.0

    def applyTo(self, matrix):
        shear_matrix = QMatrix4x4()
        shear_matrix.setToIdentity()
        if self._vertical_shear != 0.0:
            shear_matrix.setRow(1, QtGui.QVector4D(self._vertical_shear, 1.0, 0.0, 0.0))
        if self._horizontal_shear != 0.0:
            shear_matrix.setRow(0, QtGui.QVector4D(1.0, self._horizontal_shear, 0.0, 0.0))
        matrix *= shear_matrix

    def getVerticalShear(self):
        return self._vertical_shear
    
    def setVerticalShear(self, shear):
        if self._vertical_shear != shear:
            self._vertical_shear = shear
            self.update()

    def getHorizontalShear(self):
        return self._horizontal_shear

    def setHorizontalShear(self, shear):
        if self._horizontal_shear != shear:
            self._horizontal_shear = shear
            self.update()

    verticalShear = Property(float, getVerticalShear, setVerticalShear)
    horizontalShear = Property(float, getHorizontalShear, setHorizontalShear)


from PySide6.QtWidgets import QGraphicsView, QGraphicsScene, QGraphicsRectItem, QApplication
import sys

app = QApplication(sys.argv)

view = QGraphicsView()
scene = QGraphicsScene()
view.setScene(scene)

rect = QGraphicsRectItem(0, 0, 100, 100)
rect.setBrush(QtGui.Qt.GlobalColor.red)
scene.addItem(rect)

# Create and apply our custom transform
custom_transform = CustomTransform()
rect.setTransformations([custom_transform])  # Can combine with other transforms

animation = QPropertyAnimation(custom_transform, b"verticalShear")
animation.setDuration(2000)
animation.setStartValue(0.0)
animation.setEndValue(2.0)
animation.setLoopCount(-1)  # Infinite loop
animation.start()



view.show()
sys.exit(app.exec())



from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable
if TYPE_CHECKING:
    from .interface import Interface
    from .main_view import MainView
    
from PySide6 import QtWidgets, QtGui, QtCore
from pyvol_terminal.settings import utils as settings_utils
from functools import partial
import math
from pprint import pprint
import uuid
from ...settings.widgets import CenteredPopupComboBox
from dataclasses import dataclass
from ...instruments.quantities import IVOLMetrics

class settingsMixin:
    def __init__(self, *args, parentSettings=None, **kwargs):
        self.parentSettings = parentSettings  
        self._intense_interacting=False
        self._id = uuid.uuid4() 
        self.parentSettings.addIntenseInteractingChild(self.id())
        super().__init__(*args, **kwargs)
        
    
    def id(self): return self._id
    
    def processIntenseInteraction(self, state):
        if state != self._intense_interacting:
            self._intense_interacting=state
          #  self.parentSettings.updateChildInteraction(self.id(), state)

class intenseCheckComboBox(settingsMixin, QtWidgets.QComboBox):
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.activated.connect(lambda : self.processIntenseInteraction(False))

    def showPopup(self):
        self.processIntenseInteraction(True)
        return super().showPopup()
    

    
class intenseCheckMenu(settingsMixin, QtWidgets.QMenu):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.aboutToHide.connect(lambda : self.processIntenseInteraction(False))
        
    def popup(self):
        self.processIntenseInteraction(True)
        super().popup()
    
    def closeEvent2(self, event):
        super().closeEvent(event)
        self.processIntenseInteraction(False)
        return 

class RefSlotsMenu(QtWidgets.QMenu):
    def __init__(self, *args, **kwargs):
        self._connected_slots={}
        super().__init__(*args, **kwargs)
    
    def addAction(self, action: QtGui.QAction, slots: List[Callable], how) -> None:
        self._connected_slots[action.text()] = slots
        super().addAction(action)

class Settings(QtWidgets.QToolBar):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._prevAxisSettingState = None
        self.dimension_actions = {}
        self.toggle_button_actions = {}
        self._vol_source_actions=[]
        self.current_vol_configs=[]
        self._price_type_action_container = []
        self._name_idx_map = {}
        self.title_axis_dict = {"Money": "x",
                                "Expiry": "y",
                                "Volatility": "z"
                                }
        self.prev_button = 1
        self.sub_window = None
        self.vol_src_title="Toggle Vol Sources"
        self.signal_container: Dict[str, QtCore.Signal] = {}
        
        self._intense_interaction_map = {}
        self._intense_interaction = False
        
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.setToolButtonStyle(QtCore.Qt.ToolButtonTextOnly)  # Consistent style for all buttons
    
    def addIntenseInteractingChild(self, id):
        self._intense_interaction_map[id] = False
        
    def updateChildInteraction(self, id, state):
        self._intense_interaction_map[id] = state
        current_state = any(self._intense_interaction_map.values())
        if current_state != self._intense_interaction:
            self._intense_interaction = current_state
            self.processIntenseInteraction("settings", current_state)
        
    def processIntenseInteraction(self): ...
    
    def remove_settings_menu(self, title, cls):
        submenu = self._find_widget(title, cls)
        self.removeWidget(submenu)
    
    def create_vol_src_menu(self,
                            options: List[str],
                            slots: Callable | List[Callable]=None,
                            slot_args=()
                            ):
        tool_button = QtWidgets.QToolButton()
        tool_button.setText(self.vol_src_title)
        tool_button.setPopupMode(QtWidgets.QToolButton.InstantPopup)
        tool_button.setStyleSheet(settings_utils.get_settings_stylesheets("QToolButton"))
        
        menu = QtWidgets.QMenu(tool_button)
        menu.setStyleSheet(settings_utils.get_settings_stylesheets("QMenu"))
        
        action_group = QtGui.QActionGroup(menu)
        action_group.setExclusive(False)
        
        tool_button.setMenu(menu)
        tool_button.setObjectName(self.vol_src_title)
        self.addWidget(tool_button)

        for opt in options:
            action = QtGui.QAction(opt)
            action.setCheckable(True)
            action.setChecked(False)
            menu.addAction(action)
            action_group.addAction(action)
        if not slots is None:
            self.connect_vol_src_slots(action_group.actions(), slots, slot_args)
    
    def connect_vol_src_slots(self,
                              actions: List[QtGui.QAction]=None,
                              slots: List[Callable]=[],
                              slot_args=()
                              ):
        if actions is None:
            actions = self._find_widget(self.vol_src_title, QtWidgets.QToolButton).menu().actions()
        for action in actions:
            for slot in slots:
                print(f"slot_args: {slot_args}")
                action.triggered.connect(lambda checked, handler=slot, name=action.text(): handler(checked, name, *slot_args))

    def toggle_vol_src(self,
                       vol_src: str
                       ):
        tool_button = self._find_widget(self.vol_src_title, QtWidgets.QToolButton)
        for action in tool_button.menu().actions():
            if action.text() == vol_src:
                action.setChecked(True)
        
    def create_combobox_menu(self,
                             title: str,
                             options: List[str],
                             slots: Callable | List[Callable]=None,
                             slot_args=()
                             ):
        combobox = QtWidgets.QComboBox()
        combobox.setToolTip(title)
        options = options  + ["Select an option..."]
        combobox.addItems(options)
        self.addWidget(combobox)
        if not slots is None:
            self.connect_slots_to_combobox(combobox, slots, slot_args)
            
    def connect_slots_to_combobox(self,
                                  title_or_cbox: str | QtWidgets.QComboBox,
                                  slots: List[Callable|QtCore.Slot],
                                  slot_args=()
                                  ):
        if isinstance(title_or_cbox, str):
            combobox = self._find_widget(title_or_cbox, QtWidgets.QComboBox)
        else:
            combobox = title_or_cbox
            
        if combobox.count() == 0:
            combobox.setCurrentIndex(0)
        slots = [slots] if callable(slots) else slots        
        
        for s in slots:
            if not slot_args is None:
                combobox.currentTextChanged.connect(lambda text, args=slot_args, slot=s: slot(text, *args))
            else:
                combobox.currentIndexChanged.connect(s)
    
    def create_toggle_buttons(self,
                              title: str | QtWidgets.QComboBox,
                              options: List[str],
                              slots: Callable | List[Callable]=None,
                              slot_args=()
                              #toggles: List[bool]
                              ):
        tool_button = QtWidgets.QToolButton()
        tool_button.setText(title)
        tool_button.setStyleSheet(settings_utils.get_settings_stylesheets("QToolButton"))
        tool_button.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        
        menu = QtWidgets.QMenu(tool_button)
        menu.setStyleSheet(settings_utils.get_settings_stylesheets("QMenu"))

        tool_button.setPopupMode(QtWidgets.QToolButton.InstantPopup)
        tool_button.setMenu(menu)
        tool_button.setObjectName(title)
        self.addWidget(tool_button)
        
        action_group = QtGui.QActionGroup(self)
        for options in options:
            action = QtGui.QAction(options, parent=self)
            action.setCheckable(True)
            action.setChecked(False)
            #action.setChecked(toggle)
            action_group.addAction(action)
            menu.addAction(action)
        
        if not slots is None:
            self.connect_toggle_button_slots(action_group.actions(), slots, slot_args)

    def connect_toggle_button_slots(self,
                                    label_or_actions: str | List[QtGui.QAction],
                                    slots: Callable | List[Callable]=None,
                                    slot_args=()
                                    ):
        if isinstance(label_or_actions, str):
            actions = self._find_widget(label_or_actions, QtWidgets.QToolButton).menu().actions()
        else:
            actions = label_or_actions
        for action in actions:
            for slot in slots:
                action.triggered.connect(lambda check, opt=action.text(), handler=slot: handler(check, opt, *slot_args))

    def add_window_menu(self, switch_axis_call: Callable):
        tool_button = QtWidgets.QToolButton() 
        tool_button.setText("Unit Settings")
        tool_button.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        
        cls_kwargs = {"options" : self._prevAxisSettingState,
                      "slots" : [switch_axis_call],
                      "close_slots": [self._setPrevAxisSettingState],
                      "parent" : self,
                      }
        tool_button.clicked.connect(lambda: self._open_submenu(AxisParameters, cls_kwargs))
        self.addWidget(tool_button)
    
    def add_vol_src_window_menu(self, all_price_types, slots: List[Callable]):
        tool_button = QtWidgets.QToolButton()
        tool_button.setText("Display Volatility") 
        tool_button.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        cls_kwargs = {"all_price_types" : all_price_types,
                      "slots" : slots + [self._setCurrentVolSrc],
                      "current_vol_configs" : self.current_vol_configs,
                      }
        tool_button.clicked.connect(lambda: self._open_submenu(VolSelection, cls_kwargs))
        self.addWidget(tool_button)
        
    def _setCurrentVolSrc(self, vol_configs: List[IVOLMetrics]):
        self.current_vol_configs=vol_configs
            
    def _find_widget(self, name, cls):
        for submenu in self.findChildren(cls):   
            if submenu.objectName() == name:
                return submenu

    def _open_submenu(self, cls, cls_kwargs):
        geometry = self.window().geometry()
        interface_center = geometry.center()
        geometry.setSize(geometry.size() / 2)
        geometry.moveCenter(interface_center)
        
        widget = cls(**cls_kwargs)
        self.sub_window = widget
        widget.show()
        
    def _setPrevAxisSettingState(self, state):
        self._prevAxisSettingState = state
        


class BlockQLineEdit(QtWidgets.QLineEdit):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
    @QtCore.Slot(str)
    def setDisableState(self, state):
        if state == "Automatic":
            self.setDisabled(True)
        else:
            self.setDisabled(False)



class ContainerWidget(QtWidgets.QWidget):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
    def resizeEvent(self, event):
        if self.parent() and hasattr(self.window(), 'frameSize'):
            window = self.window()
            frame_size = window.frameSize()
            content_size = window.size()
            
            width_diff = frame_size.width() - content_size.width()
            height_diff = frame_size.height() - content_size.height()
            
            print(f"\ncurrent window width: {window.width()}")
            print(f"target width: {self.width() + width_diff}")
            window.resize(self.width() + width_diff,
                          self.height() + height_diff
                          )
            print(f"new width: {window.width()}")
        super().resizeEvent(event)

    
class AxisParameters(QtWidgets.QDialog):
    
    _slots = []
    
    _default_options = {"unit" : {ax : "1" for ax in "xyz"},
                        "range" : {ax : ["", ""] for ax in "xyz"},
                        "control" : {ax : "automatic" for ax in "xyz"},
                        }
        
    def __init__(self,
                 options=None,
                 slots=[],
                 close_slots=[],
                 parentSettings=None,
                 **kwargs
                 ):
        super().__init__(**kwargs)
        self.table: QtWidgets.QTableWidget = None
        self._pending_signal_args = {"unit" : {ax : None for ax in "xyz"},
                                     "range" : {ax : None for ax in "xyz"},
                                     "control" : {ax : None for ax in "xyz"}
                                     }
        if options is None:
            self._current_opt = AxisParameters._default_options
        else:
            self._current_opt = options            
        self.close_slots=close_slots
        self.parentSettings=parentSettings
        
        self._initLayout()
        self._create_unit_settings(slots)
        self._create_axis_limits_settings()
        self._create_save_exit()

        self.table.resizeRowsToContents()
        self.table.resizeColumnsToContents()
        self.adjustSize()

    def showEvent(self, arg__1):
       # self.processIntenseInteraction(True)
        return super().showEvent(arg__1)
        
    def closeEvent(self, arg__1):
        super().closeEvent(arg__1) 
       # self.processIntenseInteraction(False)
                
    def _initLayout(self):
        container = self._setupMainContainer()
        self.table = self._setupTableWidget()
        self._setupColumnSync()
        
        self.save_close_widget = QtWidgets.QWidget()
        self.save_close_widget.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        container.layout().addWidget(self.table)
        container.layout().addWidget(self.save_close_widget)
        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(container)

        self.setLayout(layout)
        

    def _setupMainContainer(self) -> ContainerWidget:
        container = ContainerWidget()
        container.setLayout(QtWidgets.QVBoxLayout())
        container.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        container.layout().setContentsMargins(0, 0, 0, 0)
        return container


    def _setupTableWidget(self) -> QtWidgets.QTableWidget:
        table = QtWidgets.QTableWidget()
        
        table.setRowCount(3)
        table.setColumnCount(2)
        
        table.setVerticalHeaderLabels([f"{ax}-axis" for ax in "xyz"])
        table.setHorizontalHeaderLabels(["Axes Units", "Axes Limits"])
        table.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        
        table.horizontalHeader().setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)
        table.horizontalHeader().setSectionResizeMode(1, QtWidgets.QHeaderView.ResizeToContents)
        table.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        table.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        return table

    def _setupColumnSync(self):
        def _sync_column_widths(logicalIndex, oldSize, newSize):
            self.table.horizontalHeader().sectionResized.disconnect(_sync_column_widths)
            other_column = 1 if logicalIndex == 0 else 0
            self.table.setColumnWidth(other_column, newSize)
            self.table.horizontalHeader().sectionResized.connect(_sync_column_widths)
        self.table.horizontalHeader().sectionResized.connect(_sync_column_widths)
        
    def _setDarkPalette(self):
        palette = self.palette()
        palette.setColor(QtGui.QPalette.Window, QtGui.QColor(53, 53, 53))
        palette.setColor(QtGui.QPalette.WindowText, QtCore.Qt.white)
        palette.setColor(QtGui.QPalette.Base, QtGui.QColor(25, 25, 25))
        palette.setColor(QtGui.QPalette.Text, QtCore.Qt.white)
        palette.setColor(QtGui.QPalette.Button, QtGui.QColor(53, 53, 53))
        palette.setColor(QtGui.QPalette.ButtonText, QtCore.Qt.white)
        self.setPalette(palette)
        
    def _create_unit_settings(self, slots):
        self._slots = slots
        
        for idx, axis in enumerate("xyz"):
            combobox = CenteredPopupComboBox()
    
            for i in range(-5,6):
                value = 10**i
                if i<0:
                    text = "0." + "0"*(-1*i - 1) + "1"
                else:
                    text =  "1" + "0"*(i)
                combobox.addItem(text)
            
            combobox.setCurrentText(self._current_opt["unit"][axis])
            
            for _ in slots:
                combobox.currentTextChanged.connect(lambda text, ax=axis: self._queue_signal(text, "unit", ax))
            self.table.setCellWidget(idx, 0, combobox)
        self.table.horizontalHeader().setStretchLastSection(True)
        
    def _queue_signal(self, value, col, axis):
        self._pending_signal_args[col].update({axis : value})
    
    def _create_save_exit(self):
        button_layout = QtWidgets.QHBoxLayout(self.save_close_widget)
        button_layout.setContentsMargins(0, 0, 0, 0)
        
        save_exit = QtWidgets.QPushButton("Save & Exit")
        exit = QtWidgets.QPushButton("Exit")
        
        save_exit.clicked.connect(self._on_save_exit)
        exit.clicked.connect(self._on_exit)
        
        button_layout.addWidget(save_exit)
        button_layout.addWidget(exit)
        
    def _on_save_exit(self):
        for col, data in self._pending_signal_args.items():
            for axis, value in data.items():
                if not value is None:
                    if value != self._current_opt[col][axis]:
                        for slot in self._slots:
                            slot(float(value), axis)
                        self._current_opt[col][axis] = value
        
        for slot in self.close_slots:
            slot(self._current_opt)
        self.window().close()
    
    def _on_exit(self):
        self.window().close()

    def _create_axis_limits_settings(self):
        for idx, axis in enumerate("xyz"):
            central_widget = QtWidgets.QWidget()
            layout = QtWidgets.QVBoxLayout(central_widget)
            
            combobox = QtWidgets.QComboBox()
            combobox.setFixedSize(combobox.sizeHint())
            
            combobox.addItems(["Automatic", "Set"])
            layout.addWidget(combobox)
            
            range_widget = QtWidgets.QWidget()
            range_layout = QtWidgets.QHBoxLayout(range_widget)
            range_layout.setContentsMargins(0, 0, 0, 0) 

            qlinedit_min = BlockQLineEdit(self._current_opt["range"][axis][0])
            if len(qlinedit_min.text()) == 0:
                qlinedit_min.setDisabled(True)
            else:
                qlinedit_min.setDisabled(False)
                
            qlinedit_max = BlockQLineEdit(self._current_opt["range"][axis][1])
            if len(qlinedit_max.text()) == 0:
                qlinedit_max.setDisabled(True)
            else:
                qlinedit_max.setDisabled(False)

            range_layout.addWidget(qlinedit_min)
            range_layout.addWidget(qlinedit_max)
            
            layout.addWidget(range_widget)
            
            combobox.setCurrentIndex(0)
            combobox.currentTextChanged.connect(qlinedit_min.setDisableState)
            combobox.currentTextChanged.connect(qlinedit_max.setDisableState)
            
            #self._queue_signal()
            
            self.table.setCellWidget(idx, 1, central_widget)
            
        self.table.horizontalHeader().setStretchLastSection(True)

class DynamicClickablePushButton(QtWidgets.QPushButton):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setEnabled(False)
        self.setStyleSheet("""
                            QPushButton:disabled {background-color: #d0d0d0;  
                            color: #666666;
                            border: 1px solid #bbbbbb;}
                            """
                            )

    def setEnabledState(self, enabled):
        self.setEnabled(enabled)
        if enabled:
            self.setStyleSheet("")
        else:
            self.setStyleSheet("""
                                QPushButton:disabled {background-color: #d0d0d0;  
                                color: #666666;
                                border: 1px solid #bbbbbb;}
                                """
                                )

class VolSelection(QtWidgets.QDialog):
    def __init__(self,
                 all_price_types: List[str],
                 current_vol_configs: List[IVOLMetrics],
                 slots,
                 *args,
                 **kwargs
                 ):
        super().__init__(*args, **kwargs)
        self.all_price_types=all_price_types
        self.current_vol_configs=current_vol_configs
        self.slots=slots
        
        self._initLayout()
        self._initOptionSelection()
        self._initCurrentVolTypes()
        self._initCommandExecution()
        
        self.show()
        self.adjustSize()
        
    def _initLayout(self):
        self.table = QtWidgets.QTableWidget()
        self.table.setRowCount(3)
        self.table.setColumnCount(2)
        self.table.setHorizontalHeaderLabels(["Display Vol", "Remove Vol"])
        self.table.setVerticalHeaderLabels(["Option Price Type", "Underlying Price Type"])
        
        self.table.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Fixed)
        self.table.setSpan(0, 1, self.table.rowCount(), 1)
        
        self.setLayout(QtWidgets.QVBoxLayout()) 
        self.layout().addWidget(self.table)
        
    def _initOptionSelection(self):
        selections = ["Make a Selection..."] + self.all_price_types
        for idx in range(2):
            combobox = CenteredPopupComboBox()
            combobox.addItems(selections)
            combobox.currentIndexChanged.connect(self._update_buttons_enable_state)
            self.table.setCellWidget(idx, 0, combobox)
            
        
         
    
    
    def _initCurrentVolTypes(self):
        """
        container_widget = QtWidgets.QWidget()
        container_layout = QtWidgets.QVBoxLayout(container_widget)
        
        for vol_selection in self.current_vol_selections:
            label = QtWidgets.QLabel(vol_selection)
            label.setSelection()
            label.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.ResizeToContents)
            container_layout.addWidget(label)
        """
        
        list_widget = QtWidgets.QListWidget()
        list_widget.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        
        def toggle_item(item):
            if item.isSelected():
                if list_widget.selectedItems() == [item]:
                    item.setSelected(False)

        list_widget.itemClicked.connect(toggle_item)
        
        self.table.setCellWidget(0, 1, list_widget)
        for vol_selection in self.current_vol_configs:
            self._addVolSelection(vol_selection)
    
    def _addVolSelection(self, vol_selection):
        QtWidgets.QListWidgetItem(vol_selection, self.table.itemAt(0, 1))
    
    def _initCommandExecution(self):
        self.create_remove = QtWidgets.QWidget()
        self.create_remove.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)

        button_layout = QtWidgets.QHBoxLayout(self.create_remove)
        button_layout.setContentsMargins(0, 0, 0, 0)
        for btn_text in ["Create", "Remove"]:
            button = DynamicClickablePushButton()
            button.setText(btn_text)
            button.clicked.connect(self._update_vol_types)
            button.clicked.connect(self._update_buttons_enable_state)
            for slot in self.slots:
                button.clicked.connect(slot)
            button_layout.addWidget(button)

    def _update_vol_types(self, text):
        if text == "Create":
            vol_config = IVOLMetrics(self.table.cellWidget(0, 0).currentText(), self.table.cellWidget(1, 0).currentText())
            self.current_vol_configs.append(vol_config)
            self.table.cellWidget(0, 0).setCurrentIndex(0), self.table.cellWidget(1, 0).setCurrentIndex(0)
            
        else:
            selected_vol_config = self.table.cellWidget(0, 1).selectedItems()
            for config in selected_vol_config.copy():
                selected_vol_config.remove(config)

    def _update_buttons_enable_state(self, text):
        button_create = self.create_remove.layout().cellWidget(0).widget()
        enable = all([self.table.cellWidget(idx, 0).currentIndex() > 0 for idx in range(2)])
        if enable != button_create.isEnabled():
            button_create.setEnabled(enable)
            
        button_remove = self.create_remove.layout().cellWidget(1).widget()
        enable = len(self.table.cellWidget(0, 1).listWidget().items()) != 0
        if enable != button_create.isEnabled():
            button_remove.setEnabled(enable)




















    def update_selection_state(self, btn_text):
        if btn_text == "Create":
            button = self.create_remove.layout().cellWidget(0).widget()
            if all([self.table.cellWidget(idx, 0).currentIndex() > 0 for idx in range(2)]):
                button.setStyleSheet("")
            else:
                button.setStyleSheet("""
                                     QPushButton:disabled {background-color: #d0d0d0;  
                                     color: #666666;
                                     border: 1px solid #bbbbbb;}
                                     """
                                     )
        elif btn_text == "Remove":
            current_selections = self.table.cellWidget(0, 1).selectedItems()
            if len(current_selections) == 0:
                button.setStyleSheet("""
                                     QPushButton:disabled {background-color: #d0d0d0;  
                                     color: #666666;
                                     border: 1px solid #bbbbbb;}
                                     """
                                     )
            else:
                self.create_remove.layout().cellWidget(0).widget()

        





import numpy as np
from dataclasses import dataclass
from typing import List, Dict, Tuple, Union, ClassVar
from pprint import pprint
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from collections import defaultdict

@dataclass(frozen=True, slots=True)
class AxisGeometry:
    __ax_to_index: ClassVar[Dict[str, int]] = {"x": 0, "y": 1, "z": 2}
    __index_to_ax: ClassVar[Dict[int, str]] = {0: "x", 1: "y", 2: "z"}
    
    @classmethod
    def getIndex(self, ax: str | int) -> int:
        return AxisGeometry.__ax_to_index.get(ax, None)
    
    @classmethod
    def getAxis(self, idx: int) -> str:
        return AxisGeometry.__index_to_ax.get(idx, None)
    
    @classmethod
    def axesIndex(cls):
        return list(AxisGeometry.__index_to_ax.keys())
    
    @classmethod
    def axesString(cls):
        return list(AxisGeometry.__ax_to_index.keys())



def verifyDirection(direction: str|int) -> int:
    if isinstance(direction, str):
        if direction not in AxisGeometry.axesIndex():
            raise ValueError("The direction is not one of x, y or z")
        else:
            direction = AxisGeometry.__ax_to_index(direction)            
    elif isinstance(direction, (int, float)) and not isinstance(direction, bool):
        if not direction in (0, 1, 2):
            raise ValueError("The direction is not one of axes indexes 0, 1, 2 (corresponds to: x, y, z)")            
    else:
        raise ValueError("The direction is not one of axes. Use one of 0, 1, 2 or x, y, z")
    return direction


def verifyAxisOrthoPoints(viewRange, axisVertexPoints):
    required_keys = {"direction", "coplanar", "perpendicular"}
    if set(axisVertexPoints) != required_keys:
        raise ValueError(f"axisOrthoPoints requires keys {required_keys}")
    
    value_to_keys = defaultdict(list)
    for key, val in axisVertexPoints.items():
        value_to_keys[val].append(key)
    
    value_to_rows = defaultdict(list)
    for i, row in enumerate(viewRange):
        for val in row:
            value_to_rows[val].append(i)
    
    overused_values = []
    for val, keys in value_to_keys.items():
        needed = len(keys)
        available = len(value_to_rows[val])
        if needed > available:
            overused_values.append((val, keys, needed, available))
    
    if overused_values:
        msgs = []
        for val, keys, needed, available in overused_values:
            keys_str = ", ".join(keys)
            msgs.append(f"value {val} used by [{keys_str}] requires {needed} distinct matches but only found {available}")
        raise ValueError("Invalid axisOrthoPoints mapping: " + " and ".join(msgs))


def verifyExtrumums(extremums):
    
    if len(extremums) != 3:
        raise ValueError(f"3 extremums are required, currently {len(extremums)}")
    
    if isinstance(extremums, dict):
        extremums_str = extremums.copy()
        extremums = 3*[None]
        for ax_str, ext in extremums_str.items():
            ax_idx = AxisGeometry.getIndex(ax_str)
            extremums[ax_idx] = ext
    
    if not all([ext in [0, 1] for ext in extremums]):
        raise ValueError("extremum values must be either 0 or 1")
    return extremums


def verifyPlaneDirection(direction, planeDirection):
    if planeDirection is not None and direction == planeDirection:
        raise ValueError(f"planeDirection must be different to direction")

    if not planeDirection in AxisGeometry.axesIndex() + AxisGeometry.axesString():
        raise ValueError(f"planeDirection must be either one of (0, 1, 2) or ('x', 'y', 'z')")
    
    if isinstance(planeDirection, str):
        planeDirection = AxisGeometry.getIndex(planeDirection)
        
    return planeDirection


def verifyVertexPointsDirection(viewRange, direction, vertexPoints) -> int:
    diff_ortho_vertices = np.array(vertexPoints["planeDirection"]) - np.array(vertexPoints["perpendicularDirection"])
    
    if  np.sum(diff_ortho_vertices == 0) > 1:
        raise ValueError(f"planeDirection and perpendicularDirection: {vertexPoints} are not orthogonal to each other")
    
    directionFromVertices = diff_ortho_vertices[diff_ortho_vertices == 0].astype(int)
    
    if direction is not None and direction != directionFromVertices:
        raise ValueError(f"The given axis-direction {direction} is not the axis for the axisVertexPoints")
    
    if direction is None:
        direction = directionFromVertices
    
    for ax, point in vertexPoints.items():
        for idx, p in enumerate(point):
            lim = viewRange[idx]
            if not p in lim:
                raise ValueError(f"Vertex direction: {ax} with point {point} is not in viewRange {lim}")
    
    return direction
                
                


from __future__ import annotations
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable
if TYPE_CHECKING:
    from pyvol_terminal.windows import InterfaceWindow
    from ...instruments.utils import InstrumentManager
    from ...instruments.instruments import ABCInstrument, Option
    from .utils import ExpiryAttributes
    
    

from PySide6 import QtWidgets, QtCore, QtGui
import numpy as np
from . import stylesheets as stylesheets, table_builder, extra_widgets, settings
from datetime import datetime
from .extra_widgets import OptionExpiryCellItem, OptionNameCellItem, OptionMetricsWithTabs
from . import extra_widgets
from ...quantities import engines
from . import utils as utils_omon
from .. import custom_widgets
from ...metric_attributes import get_metric_category

class OptionTableWidget(QtWidgets.QTableWidget):
    focusOutSig = QtCore.Signal()
    
    def __init__(self,
                 instrument_manager: InstrumentManager,
                 metricLabels: List[str],
                 expiry_attributes: ExpiryAttributes,
                 n_rows,
                 option_type,
                 expiry_underlying_map=None,
                 tick_label_engine=None,
                 **kwargs
                 ):
        super().__init__(**kwargs)
        self.options_container: Dict[str, Option]=instrument_manager.options_instrument_container.objects
        self.options_maps = instrument_manager.options_instrument_container.maps
        self.name=option_type
        self._right_clicked=False
        self.metrics = [metric.lower() for metric in metricLabels]
        self.expiry_attributes: Dict[float, ExpiryAttributes] = expiry_attributes
        self.option_type=option_type
        self.visible_instruments: List[str]=[]
        self.subtable_instruments: Dict[str, List[QtWidgets.QTableWidget]]={}
        self.subtable_widget_container: List[extra_widgets.OptionMetricsWithTabs]=[]
        self.idx_instrument_map={}
        self.right_click_menu = extra_widgets.RightClickMenu(parent=self)
        self.expiry_underlying_map: Dict[float, ABCInstrument]=expiry_underlying_map
        self._expiry_cells: set = set()
        self._cell_expiry_map: Dict[int, float] = {}
        
        self.focusOutSig.connect(self.clearSelection)
        self.focusOutSig.connect(self.right_click_menu.close)
        
        if tick_label_engine is None:
            self.tick_label_engine = engines.TickEngineManager("Strike", "Expiry", "Implied Volatility")
        
        column_labels = ['Instrument Name'] + metricLabels        
        
        self.block_update=False
        self.last_values={ticker : {metric : None for metric in self.metrics} for ticker in self.options_container}
        
        self.initTableDimensions(n_rows, column_labels)        
        self.initTableContents()
        self.initTableStyle()
        self.initTableLayout()
    
    def add_expiry_underlying_map(self, expiry_underlying_map):
        self.expiry_underlying_map=expiry_underlying_map
        self.update_expiry_cell_item=True
    
    def initTableLayout(self):        
        for jdx in range(self.columnCount()):
            if jdx == 0:
                self.horizontalHeader().setSectionResizeMode(jdx, QtWidgets.QHeaderView.Stretch)  
            else:
                self.horizontalHeader().setSectionResizeMode(jdx, QtWidgets.QHeaderView.ResizeToContents)
                
        self.autoFillBackground()
        self.horizontalHeader().setHighlightSections(False)
        self.horizontalHeader().setSectionsMovable(False)
        self.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
        
        expiry_cells=[]
        self._cell_expiry_map={}
        for expiry_attribute in self.expiry_attributes.values():
            expiry_cells.append(expiry_attribute.expiry_idx)
            self._cell_expiry_map[expiry_attribute.expiry_idx] = expiry_attribute.expiry
        
        self._expiry_cells = set(expiry_cells)
        
    def currentColumns(self):
        return [self.horizontalHeaderItem(c).text() for c in range(self.columnCount())]

    def initTableDimensions(self, n_rows, column_labels):
        self.verticalHeader().setVisible(False)
        self.horizontalHeader().setVisible(True)
        
        self.setRowCount(n_rows)
        self.setColumnCount(len(column_labels))
        
        for idx, col in enumerate(column_labels):
            col_item = extra_widgets.TableColumnItem(col)
            if idx == 0:
                col_item.setTextAlignment(QtCore.Qt.AlignLeft)
                font = QtGui.QFont("Neue Haas Grotesk", 12)
                col_item.setFont(font)
            self.setHorizontalHeaderItem(idx, col_item)
    
    def initTableStyle(self):
        self.setShowGrid(True)
        self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        
        self.setItemDelegate(extra_widgets.CustomDelegate(self))
        self.setStyleSheet(stylesheets.get_settings_stylesheets("QTableWidget"))   
        self.horizontalHeader().setSectionsClickable(True)   
        
        self.itemSelectionChanged.connect(self._select_subtable)    
                
    def initTableContents(self):
        table_widget_items=[]
        self.exp_container = extra_widgets.CellItemContainer()
        for expiry, expiry_attribute in self.expiry_attributes.items():
            instrument_names = list(self.options_maps.type_expiry_strike_instrument_map[self.option_type][expiry].values())

            for strike, ticker in self.options_maps.type_expiry_strike_instrument_map[self.option_type][expiry].items():
                self.idx_instrument_map[expiry_attribute.strike_idx_map[strike]] = ticker

            table_widget_items = table_builder.create_expiry_section(expiry,
                                                                    expiry_attribute.strikes, 
                                                                    instrument_names, 
                                                                    self.metrics,
                                                                    expiry_attribute.expiry_idx
                                                                    )     
            self.exp_container.addItem(table_widget_items[0]["item"]) 
            for table_widget_item_dict in table_widget_items:
                pos, item = table_widget_item_dict["pos"], table_widget_item_dict["item"]
                self.setItem(*pos, item)
                if pos[0] == expiry_attribute.expiry_idx:
                    self.setRowHidden(pos[0], False)
                else:
                    self.setRowHidden(pos[0], True)
                
    def check_text_width(self, table_widget_items: List[Dict[str, Tuple|OptionExpiryCellItem]]):
        shrink_text={}
        for table_widget_item_dict in table_widget_items:
            item = table_widget_item_dict["item"]
            t = type(item)
            if isinstance(t, (OptionExpiryCellItem, OptionNameCellItem)):
                if not t in shrink_text:
                    shrink_text[t] = False
                if item.does_text_fit():
                    shrink_text[t] = True
        for table_widget_item_dict in table_widget_items:
            item = table_widget_item_dict["item"]
            t = type(item)
            if t in shrink_text:
                if shrink_text[t]:
                    new_pointSize=item.font().pointSize()-1
                    item.font().setPointSize(new_pointSize)

    def setRowHidden(self, row, hide):
        if hide:
            if row in self.idx_instrument_map:
                ticker = self.idx_instrument_map[row]
                if ticker in self.visible_instruments:
                    self.visible_instruments.remove(ticker)
        else:
            if row in self.idx_instrument_map:
                ticker = self.idx_instrument_map[row]
                if not ticker in self.visible_instruments:
                    self.visible_instruments.append(ticker)
        return super().setRowHidden(row, hide)
    
    def updateFromOption(self, option_object: Option):
        exp = option_object.expiry
        expiry_attr_dataclass = self.expiry_attributes[exp]
        idx = expiry_attr_dataclass.strike_idx_map[option_object.strike]
        
        for jdx, metric in enumerate(self.metrics, start=1):
            value = getattr(option_object, metric)
            if value == self.last_values[option_object.ticker][metric]:
                continue
            else:
                self.last_values[option_object.ticker][metric]=value
            
            text_item = self.item(idx, jdx)
            
            if get_metric_category(metric) == "volatility":     
                if value == value:            
                    if not option_object.OTM:
                        value_str = f"{(np.round(value, 2))}*"
                    else:
                        value_str = str(np.round(value, 2))
                else:
                    continue
            else:
                value_str = f"{value:,.2f}"
            text_item.setText(value_str)  

    def _internalTableUpdate(self, table: OptionMetricsWithTabs, instrument_names: List[str]):
        for ticker in instrument_names:
            option_object=self.options_container[ticker]
            table.updateFromOption(option_object)
    
    def update_table(self):
        if not self.block_update:
            self.setItemDelegate(None)
            self._internalTableUpdate(self, self.visible_instruments)
                
            if not self.expiry_underlying_map is None:
                for expiry, expiry_attr in self.expiry_attributes.items():
                    underlying_object = self.expiry_underlying_map[expiry]
                    idx = expiry_attr.expiry_idx
                    text_item = self.item(idx, 0)
                    text_item.update_from_base(underlying_object.mid)
            
            for table_widget in self.subtable_widget_container:
                self._internalTableUpdate(table_widget, table_widget.instrument_names)
            self.setItemDelegate(extra_widgets.CustomDelegate(self))  
                
    def _select_subtable(self):
        selected_items = self.selectedItems()
        if not selected_items:
            return
        self.blockSignals(True)
        for item in selected_items:
            row, col = item.row(), item.column()
            if row in self._expiry_cells:
                expiry = self._cell_expiry_map[row]
                expiry_attr = self.expiry_attributes[expiry]
                
                if col == 0:
                    for j in range(self.columnCount()):
                        self.item(row, j).setSelected(True)
                    for strike_row in expiry_attr.strike_idx_map.values():
                        if not self.isRowHidden(strike_row):
                            for j in range(self.columnCount()):
                                if self.item(strike_row, j):
                                    self.item(strike_row, j).setSelected(True)
                else:
                    for strike_row in expiry_attr.strike_idx_map.values():
                        if not self.isRowHidden(strike_row) and self.item(strike_row, col):
                            self.item(strike_row, col).setSelected(True)
            else:
                if col == 0:
                    for j in range(self.columnCount()):
                        if self.item(row, j):
                            self.item(row, j).setSelected(True)
        
        self.blockSignals(False)

    def reset_right_click_flag(self):
        self._right_clicked = False
        
    def mouseMoveEvent(self, event):
        if event.buttons() & QtCore.Qt.LeftButton:
            if self.right_click_menu.isVisible():
                self.right_click_menu.hide()
        super().mouseMoveEvent(event)
    
    def focusOutEvent(self, event):
        self.focusOutSig.emit()
        return super().focusOutEvent(event)
    
    def keyPressEvent(self, event: QtGui.QKeyEvent):
        if event.key() == QtGui.Qt.Key_Escape and event.modifiers() == QtGui.Qt.NoModifier:
            self.focusOutSig.emit()
        return super().keyPressEvent(event)
            
    def mousePressEvent(self, event: QtGui.QMouseEvent):
        if event.button() == QtCore.Qt.LeftButton:
            self._right_clicked = False
            super().mousePressEvent(event)
        
        if event.button() == QtCore.Qt.RightButton:
            self._right_clicked = True
            rows = [item.row() for item in self.selectedItems()]
            cols = [item.column() for item in self.selectedItems()]
            name_metric_map={}
            for row, col in zip(rows, cols):
                if 0 < col < 1 + len(self.metrics):
                    if row in self._expiry_cells:
                        continue
                    ticker = self.idx_instrument_map[row]
                    selected_column = self.currentColumns()[col]
                    if get_metric_category(selected_column.lower()) == "price":
                        
                    
                        if not ticker in name_metric_map:
                            inner_dict = {"object" : self.options_container[ticker],
                                          "selectedPriceTypes" : {"label" : [selected_column],
                                                                  "attribute" : [selected_column.lower()]}
                                        }
                            name_metric_map[ticker] = inner_dict
                        else:
                            name_metric_map[ticker]["selectedPriceTypes"]["label"].append(selected_column)
                            name_metric_map[ticker]["selectedPriceTypes"]["attribute"].append(selected_column.lower())

            if len(name_metric_map) > 0:
                self.right_click_menu.popup(event.globalPos(), name_metric_map, self) 
                
        super().mousePressEvent(event) 
    
    @QtCore.Slot(QtCore.QPoint, dict)
    def create_subtable(self, q_point, name_metric_map):
        app = QtWidgets.QApplication.instance()
    
        getattr_map = {"ivol" : "get_ivol",
                       "delta" : "get_delta",
                       "gamma" : "get_gamma",
                       "vega" : "get_vega",
                       "theta" : "get_theta",
                       "rho" : "get_rho",
                       }

        metric_table = extra_widgets.OptionMetricsWithTabs(name_metric_map=name_metric_map, getattr_map=getattr_map, parent=self)        
        geomtry = QtCore.QRect(int(q_point.x()), int(q_point.y()), int(metric_table.width()), int(metric_table.height()))
        self.subtable_widget_container.append(metric_table)
        metric_table.show()
        #################app.open_sub_window(metric_table, geomtry, True, parent=self)
        
    @QtCore.Slot()
    def closingSubtle(self, table):
        self.subtable_widget_container.remove(table)
        

class StrikeTableColumn(QtWidgets.QTableWidget):
    sigRowsUpdated = QtCore.Signal(object)
    def __init__(self,
                 n_rows: int,
                 expiry_attributes: Dict[str, ExpiryAttributes]
                 ):
        super().__init__()
        self.name="strike"
        self.expiry_attributes=expiry_attributes
        self.comboboxes: List[extra_widgets.StrikeOptionsComboBox]=[]
        self.current_expiry_strike_map={expiry : [] for expiry in self.expiry_attributes}
        self.strike_center=None
        self.row_hidden_callbacks={idx : [] for idx in range(n_rows)}
        self.max_width=0
        self.init_layout(n_rows)
        self.setStyleSheet(stylesheets.get_settings_stylesheets("QTableWidget"))
        self.autoFillBackground()
        
    def init_layout(self, n_rows):
        self.setColumnCount(1)
        self.setRowCount(n_rows)

        for expiry_attribute in self.expiry_attributes.values():
            select_strikes_combobox = extra_widgets.StrikeOptionsComboBox()
            select_strikes_combobox.addItem(str(0))

            self.setCellWidget(expiry_attribute.expiry_idx, 0, select_strikes_combobox)
            
            select_strikes_combobox.blockSignals(True)
            select_strikes_combobox.setCurrentText(str(0))    
            select_strikes_combobox.currentTextChanged.connect(lambda selected_text, exp_attr=expiry_attribute: self.update_strikes(selected_text, exp_attr))
            select_strikes_combobox.blockSignals(False)
            self.comboboxes.append(select_strikes_combobox)
            for idx, strike in enumerate(expiry_attribute.strikes, start=1):
                text_widget_item = extra_widgets.StrikeCellItem(str(strike))
                select_strikes_combobox.addItem(str(idx))
                self.setItem(expiry_attribute.strike_idx_map[strike], 0, text_widget_item)
                
        for idx in range(self.rowCount()):
            self.setRowHidden(idx, False)
        self.horizontalHeader().setSectionResizeMode(0, QtWidgets.QHeaderView.ResizeToContents)
        self.max_width = self.columnWidth(0)
        for idx in range(self.rowCount()):
               self.setRowHidden(idx, True)
        for expiry_attribute in self.expiry_attributes.values():
            self.setRowHidden(expiry_attribute.expiry_idx, False)
                    
        self.horizontalHeader().setSectionResizeMode(0, QtWidgets.QHeaderView.Fixed)
        self.columnWidth(self.max_width)

        self.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
        self.horizontalHeader().setVisible(True)
        self.setHorizontalHeaderItem(0, extra_widgets.TableColumnItem(""))
        self.verticalHeader().setVisible(False)
        
    def add_row_hidden_callback(self, callback, idx):
        self.row_hidden_callbacks[idx].append(callback)

    def setRowHidden(self, idx, hide_flag):
        super().setRowHidden(idx, hide_flag)
        if idx in self.row_hidden_callbacks:
            for callback in self.row_hidden_callbacks[idx]:
                callback(idx, hide_flag)
        self.sigRowsUpdated.emit(self)
    
    def set_strikes(self, new_n_strikes, expiry_attribute):
        old_strikes = self.current_expiry_strike_map[expiry_attribute.expiry]
        new_strikes, _, _  = utils_omon.get_closest_n_strikes(self.strike_center, expiry_attribute.strikes, new_n_strikes)
        self.current_expiry_strike_map[expiry_attribute.expiry] = new_strikes

        if len(new_strikes) > len(old_strikes):
            strikes_to_change = [strike for strike in new_strikes if not strike in old_strikes]
            self.update_table_rows(expiry_attribute, strikes_to_change, False)
        elif len(new_strikes) == len(old_strikes):
            strikes_to_add = [strike for strike in new_strikes if not strike in old_strikes]
            self.update_table_rows(expiry_attribute, strikes_to_add, False)
            strikes_to_remove = [strike for strike in old_strikes if not strike in new_strikes]
            self.update_table_rows(expiry_attribute, strikes_to_remove, True)
        else:
            strikes_to_change = [strike for strike in old_strikes if not strike in new_strikes]
            self.update_table_rows(expiry_attribute, strikes_to_change, True)

    def update_strikes(self, new_n_strikes, expiry_attribute):
        new_n_strikes = int(float(new_n_strikes))
        if new_n_strikes == len(self.current_expiry_strike_map[expiry_attribute.expiry]):
            return
        else:
            self.set_strikes(new_n_strikes, expiry_attribute)

    def update_table_rows(self, expiry_attribute: ExpiryAttributes, strikes_to_change: List[float], hide_flag: bool):
        for strike in strikes_to_change:
            idx = expiry_attribute.strike_idx_map[strike]
            self.setRowHidden(idx, hide_flag)

    def change_center(self, center):
        self.strike_center = float(center)
        for expiry, expiry_attribute in self.expiry_attributes.items():
            if len(self.current_expiry_strike_map[expiry]) == 0:
                continue
            else:
                self.set_strikes(len(self.current_expiry_strike_map[expiry]), expiry_attribute)
                
    def bulk_change_strike_num(self, new_n_strike):
        for combobox in self.comboboxes:
            if new_n_strike > combobox.maxCount():
                combobox.setCurrentIndex(combobox.maxCount() - 1)
            else:
                combobox.setCurrentText(str(new_n_strike))

